# Copyright (c) 2018 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import logging
import time
from pathlib import Path
from typing import Callable, Iterable, Type

from . import _parser
from ._data_type_builder import DataTypeBuilder, UndefinedDataTypeError
from ._dsdl import DefinitionVisitor, ReadableDSDLFile
from ._error import FrontendError, InternalError, InvalidDefinitionError
from ._serializable import CompositeType, Version

_logger = logging.getLogger(__name__)


class FileNameFormatError(InvalidDefinitionError):
    """
    Raised when a DSDL definition file is named incorrectly.
    """

    def __init__(self, text: str, path: Path):
        super().__init__(text=text, path=Path(path))


class PathInferenceError(UndefinedDataTypeError):
    """
    Raised when the namespace, type, fixed port ID, or version cannot be inferred from a file path.
    """

    def __init__(self, text: str, dsdl_path: Path, valid_dsdl_roots: list[Path]):
        super().__init__(text=text, path=Path(dsdl_path))
        self.valid_dsdl_roots = valid_dsdl_roots[:] if valid_dsdl_roots is not None else None


class DSDLDefinition(ReadableDSDLFile):
    """
    A DSDL type definition source abstracts the filesystem level details away, presenting a higher-level
    interface that operates solely on the level of type names, namespaces, fixed identifiers, and so on.
    Upper layers that operate on top of this abstraction do not concern themselves with the file system at all.

    :param file_path: The path to the DSDL file.
    :param root_namespace_path: The path to the root namespace directory. `file_path` must be a descendant of this path.
                                See `from_first_in` for a more flexible way to create a DSDLDefinition object.
    :raises InvalidDefinitionError: If file_path does not exist.
    """

    @classmethod
    def _infer_path_to_root_from_first_found(cls, dsdl_path: Path, valid_dsdl_roots: list[Path]) -> Path:
        """
        See `from_first_in` for documentation on this logic.
        :return The path to the root namespace directory.
        """
        if valid_dsdl_roots is None:
            raise ValueError("valid_dsdl_roots was None")

        # INFERENCE 1: The easiest inference is when the target path is relative to the current working directory and
        # the root is a direct child folder. In this case we allow targets to be specified as simple, relative paths
        # where we infer the root from the first part of each path.
        if len(valid_dsdl_roots) == 0:
            # if we have no valid roots we can only infer the root of the path.
            if dsdl_path.is_absolute():
                # but if the path is absolute we refuse to infer the root as this would cause us to search the entire
                # filesystem for DSDL files and it's almost certainly wrong.
                raise PathInferenceError(
                    f"No valid roots provided for absolute path {str(dsdl_path)}. Unable to infer root without "
                    "more information.",
                    dsdl_path,
                    valid_dsdl_roots,
                )
            else:
                # if the path is relative we'll assume the root is the top-most folder in the path.
                directly_inferred = Path(dsdl_path.parts[0])
                try:
                    directly_inferred.resolve(strict=True)
                except FileNotFoundError:
                    raise PathInferenceError(
                        f"No valid root found in path {str(dsdl_path)} and the inferred root {str(directly_inferred)} "
                        "does not exist. You either need to change your working directory to the folder that contains "
                        "this root folder or provide a valid root path.",
                        dsdl_path,
                        valid_dsdl_roots,
                    )
                return directly_inferred

        # INFERENCE 2: The next easiest inference is when the target path is relative to a known dsdl root. These
        # operations should work with pure paths and not require filesystem access.
        resolved_dsdl_path = dsdl_path.resolve(strict=False) if dsdl_path.is_absolute() else None
        for path_to_root in valid_dsdl_roots:
            # First we try the paths as-is...
            try:
                _ = dsdl_path.relative_to(path_to_root)
            except ValueError:
                pass
            else:
                return path_to_root
            # then we try resolving the root path if it is absolute
            if path_to_root.is_absolute() and resolved_dsdl_path is not None:
                path_to_root_resolved = path_to_root.resolve(strict=False)
                try:
                    _ = resolved_dsdl_path.relative_to(path_to_root_resolved).parent
                except ValueError:
                    pass
                else:
                    return path_to_root_resolved

        # INFERENCE 3: If the target is relative then we can try to find a valid root by looking for the file in the
        # root directories. This is a stronger inference than the previous one because it requires the file to exist
        # but we do it second because it reads the filesystem.
        if not dsdl_path.is_absolute():
            for path_to_root in valid_dsdl_roots:
                path_to_root_parent = path_to_root
                while path_to_root_parent != path_to_root_parent.parent:
                    # Weld together and check only if the root's last part is the same name as the target's first part.
                    # yes:
                    #     path/to/root + root/then/Type.1.0.dsdl <- /root == root/
                    # no:
                    #     path/to/not_root + root/then/Type.1.0.dsdl <- /not_root != root/
                    if (
                        path_to_root_parent.parts[-1] == dsdl_path.parts[0]
                        and (path_to_root_parent.parent / dsdl_path).exists()
                    ):
                        return path_to_root_parent
                    path_to_root_parent = path_to_root_parent.parent

        # INFERENCE 4: A weaker, but valid inference is when the target path is a child of a known root folder name.
        # This is only allowed if dsdl roots are top-level namespace names and not paths.
        root_parts = [x.parts[-1] for x in valid_dsdl_roots if len(x.parts) == 1]
        parts = list(dsdl_path.parent.parts)
        for i, part in list(enumerate(parts)):
            if part in root_parts:
                return Path().joinpath(*parts[: i + 1])
                # +1 to include the root folder
        raise PathInferenceError(f"No valid root found in path {str(dsdl_path)}", dsdl_path, valid_dsdl_roots)

    @classmethod
    def from_first_in(cls: Type["DSDLDefinition"], dsdl_path: Path, valid_dsdl_roots: list[Path]) -> "DSDLDefinition":
        """
        Creates a DSDLDefinition object by inferring the path to the namespace root of a DSDL file given a set
        of valid roots and, if the dsdl path is relative, resolving the dsdl path relative to said roots. The logic used
        prefers an instance of `dsdl_path` found to exist under a valid root but will degrade to pure-path string
        matching if no file is found (If this does not yield a valid path to an existing dsdl file an exception is
        raised). Because this logic uses the first root path that passes one of these two inferences the order of the
        valid_dsdl_roots list matters.

        :param dsdl_path:           The path to the alleged DSDL file.
        :param valid_dsdl_roots:    The ordered set of valid root names or paths under which the type must reside.
                                    This argument is accepted as a list for ordering but no de-duplication is performed
                                    as the caller is expected to provide a correct set of paths.
        :return A new DSDLDefinition object
        :raises PathInferenceError: If the namespace root cannot be inferred from the provided information.
        :raises InvalidDefinitionError: If the file does not exist.
        """
        root_path = cls._infer_path_to_root_from_first_found(dsdl_path, valid_dsdl_roots)
        if not dsdl_path.is_absolute():
            dsdl_path_resolved = (root_path.parent / dsdl_path).resolve(strict=False)
        else:
            dsdl_path_resolved = dsdl_path.resolve(strict=False)
        return cls(dsdl_path_resolved, root_path)

    def __init__(self, file_path: Path, root_namespace_path: Path):
        """ """
        # Normalizing the path and reading the definition text
        self._file_path = Path(file_path)
        del file_path

        if not self._file_path.exists():
            raise InvalidDefinitionError(
                "Attempt to construct ReadableDSDLFile object for file that doesn't exist.", self._file_path
            )

        self._root_namespace_path = Path(root_namespace_path)
        del root_namespace_path
        self._text: str | None = None

        # Checking the sanity of the root directory path - can't contain separators
        if CompositeType.NAME_COMPONENT_SEPARATOR in self._root_namespace_path.name:
            raise FileNameFormatError("Invalid namespace name", path=self._root_namespace_path)

        relative_path = self._root_namespace_path.name / self._file_path.relative_to(self._root_namespace_path)

        # Parsing the basename, e.g., 434.GetTransportStatistics.0.1.dsdl
        basename_components = relative_path.name.split(".")[:-1]
        str_fixed_port_id: str | None = None
        if len(basename_components) == 4:
            str_fixed_port_id, short_name, str_major_version, str_minor_version = basename_components
        elif len(basename_components) == 3:
            short_name, str_major_version, str_minor_version = basename_components
        else:
            raise FileNameFormatError("Invalid file name", path=self._file_path)

        # Parsing the fixed port ID, if specified; None if not
        if str_fixed_port_id is not None:
            try:
                self._fixed_port_id: int | None = int(str_fixed_port_id)
            except ValueError:
                raise FileNameFormatError(
                    "Not a valid fixed port-ID: %s. "
                    "Namespaces are defined as directories; putting the namespace name in the file name will not work. "
                    'For example: "foo/Bar.1.0.dsdl" is OK (where "foo" is a directory); "foo.Bar.1.0.dsdl" is not.'
                    % str_fixed_port_id,
                    path=self._file_path,
                ) from None
        else:
            self._fixed_port_id = None

        # Parsing the version numbers
        try:
            self._version = Version(major=int(str_major_version), minor=int(str_minor_version))
        except ValueError:
            raise FileNameFormatError("Could not parse the version numbers", path=self._file_path) from None

        # Finally, constructing the name
        namespace_components = list(relative_path.parent.parts)
        for nc in namespace_components:
            if CompositeType.NAME_COMPONENT_SEPARATOR in nc:
                raise FileNameFormatError(f"Invalid name for namespace component: {nc!r}", path=self._file_path)
        self._name: str = CompositeType.NAME_COMPONENT_SEPARATOR.join(namespace_components + [str(short_name)])

        self._cached_type: CompositeType | None = None

    # +-----------------------------------------------------------------------+
    # | ReadableDSDLFile :: INTERFACE                                         |
    # +-----------------------------------------------------------------------+
    @property
    def file_path(self) -> Path:
        return self._file_path

    def read(
        self,
        lookup_definitions: Iterable[ReadableDSDLFile],
        definition_visitors: Iterable[DefinitionVisitor],
        print_output_handler: Callable[[int, str], None],
        allow_unregulated_fixed_port_id: bool,
    ) -> CompositeType:
        log_prefix = "%s.%d.%d" % (self.full_name, self.version.major, self.version.minor)
        if self._cached_type is not None:
            _logger.debug("%s: Cache hit", log_prefix)
            return self._cached_type

        started_at = time.monotonic()

        # Remove the target definition from the lookup list in order to prevent
        # infinite recursion on self-referential definitions.
        lookup_definitions = list(filter(lambda d: d != self, lookup_definitions))

        _logger.debug(
            "%s: Starting processing with %d lookup definitions located in root namespaces: %s",
            log_prefix,
            len(lookup_definitions),
            ", ".join(set(sorted(map(lambda x: x.root_namespace, lookup_definitions)))),
        )
        try:
            builder = DataTypeBuilder(
                definition=self,
                lookup_definitions=lookup_definitions,
                definition_visitors=definition_visitors,
                print_output_handler=print_output_handler,
                allow_unregulated_fixed_port_id=allow_unregulated_fixed_port_id,
            )

            _parser.parse(self.text, builder)

            self._cached_type = builder.finalize()
            _logger.info(
                "%s: Processed in %.0f ms; category: %s, fixed port ID: %s",
                log_prefix,
                (time.monotonic() - started_at) * 1e3,
                type(self._cached_type).__name__,
                self._cached_type.fixed_port_id,
            )
            return self._cached_type
        except FrontendError as ex:  # pragma: no cover
            ex.set_error_location_if_unknown(path=self.file_path)
            raise ex
        except (MemoryError, SystemError):  # pragma: no cover
            raise
        except Exception as ex:  # pragma: no cover
            raise InternalError(culprit=ex, path=self.file_path) from ex

    # +-----------------------------------------------------------------------+
    # | DSDLFile :: INTERFACE                                                 |
    # +-----------------------------------------------------------------------+
    @property
    def composite_type(self) -> CompositeType | None:
        return self._cached_type

    @property
    def full_name(self) -> str:
        return self._name

    @property
    def name_components(self) -> list[str]:
        return self._name.split(CompositeType.NAME_COMPONENT_SEPARATOR)

    @property
    def short_name(self) -> str:
        return self.name_components[-1]

    @property
    def full_namespace(self) -> str:
        return str(CompositeType.NAME_COMPONENT_SEPARATOR.join(self.name_components[:-1]))

    @property
    def root_namespace(self) -> str:
        return self.name_components[0]

    @property
    def text(self) -> str:
        if self._text is None:
            with open(self._file_path) as f:
                self._text = str(f.read())
        return self._text

    @property
    def version(self) -> Version:
        return self._version

    @property
    def fixed_port_id(self) -> int | None:
        return self._fixed_port_id

    @property
    def has_fixed_port_id(self) -> bool:
        return self.fixed_port_id is not None

    @property
    def root_namespace_path(self) -> Path:
        return self._root_namespace_path

    # +-----------------------------------------------------------------------+
    # | Python :: SPECIAL FUNCTIONS                                           |
    # +-----------------------------------------------------------------------+
    def __hash__(self) -> int:
        return hash((self.full_name, self.version))

    def __eq__(self, other: object) -> bool:
        """
        Two definitions will compare equal if they share the same name AND version number.
        Definitions of the same name but different versions are not considered equal.
        """
        if isinstance(other, DSDLDefinition):
            return self.full_name == other.full_name and self.version == other.version
        return NotImplemented  # pragma: no cover

    def __str__(self) -> str:
        try:
            return "DSDLDefinition(full_name=%r, version=%r, fixed_port_id=%r, file_path=%s)" % (
                self.full_name,
                self.version,
                self.fixed_port_id,
                self.file_path,
            )
        except AttributeError:  # pragma: no cover
            return "DSDLDefinition(UNINITIALIZED)"

    __repr__ = __str__


# +-[UNIT TESTS]------------------------------------------------------------------------------------------------------+


def _unittest_dsdl_definition_read_non_existent() -> None:
    from pytest import raises as expect_raises

    target = Path("root", "ns", "Target.1.1.dsdl")
    with expect_raises(InvalidDefinitionError):
        _ = DSDLDefinition(target, target.parent)


def _unittest_dsdl_definition_read_text(temp_dsdl_factory) -> None:  # type: ignore
    from pytest import raises as expect_raises

    target_root = Path("root", "ns")
    target_file_path = Path(target_root / "Target.1.1.dsdl")
    dsdl_file = temp_dsdl_factory.new_file(target_root / target_file_path, "@sealed")
    with expect_raises(ValueError):
        _target_definition = DSDLDefinition(dsdl_file, target_root)
        # we test first that we can't create the object until we have a target_root that contains the dsdl_file

    target_definition = DSDLDefinition(dsdl_file, dsdl_file.parent.parent)
    assert "@sealed" == target_definition.text


def _unittest_type_from_path_inference() -> None:
    from pytest import raises as expect_raises

    # pylint: disable=protected-access

    dsdl_file = Path("/repo/uavcan/foo/bar/435.baz.1.0.dsdl").resolve()
    path_to_root = DSDLDefinition._infer_path_to_root_from_first_found(dsdl_file, [Path("/repo/uavcan").resolve()])
    namespace_parts = dsdl_file.parent.relative_to(path_to_root.parent).parts

    assert path_to_root == Path("/repo/uavcan").resolve()
    assert namespace_parts == ("uavcan", "foo", "bar")

    # The simplest inference made is when relative dsdl paths are provided with no additional information. In this
    # case the method assumes that the relative path is the correct and complete namespace of the type:

    # relative path
    root = DSDLDefinition._infer_path_to_root_from_first_found(
        Path("uavcan/foo/bar/435.baz.1.0.dsdl"), [Path("uavcan")]
    )
    assert root == Path("uavcan")

    with expect_raises(ValueError):
        _ = DSDLDefinition._infer_path_to_root_from_first_found(
            Path("/repo/uavcan/foo/bar/435.baz.1.0.dsdl").resolve(), None  # type: ignore
        )

    # If an absolute path is provided along with a path-to-root "hint" then the former must be relative to the
    # latter:

    # dsdl file path is not contained within the root path
    with expect_raises(PathInferenceError):
        _ = DSDLDefinition._infer_path_to_root_from_first_found(
            Path("/repo/uavcan/foo/bar/435.baz.1.0.dsdl").resolve(), [Path("/not-a-repo").resolve()]
        )

    root = DSDLDefinition._infer_path_to_root_from_first_found(
        Path("/repo/uavcan/foo/bar/435.baz.1.0.dsdl").resolve(), [Path("/repo/uavcan").resolve()]
    )
    assert root == Path("/repo/uavcan").resolve()

    # The priority is given to paths that are relative to the root when both simple root names and paths are provided:
    root = DSDLDefinition._infer_path_to_root_from_first_found(
        Path("/repo/uavcan/foo/bar/435.baz.1.0.dsdl").resolve(), [Path("foo"), Path("/repo/uavcan").resolve()]
    )
    assert root == Path("/repo/uavcan").resolve()

    root = DSDLDefinition._infer_path_to_root_from_first_found(
        Path("repo/uavcan/foo/bar/435.baz.1.0.dsdl"), [Path("foo"), Path("repo/uavcan")]
    )
    assert root == Path("repo/uavcan")

    # Finally, the method will infer the root namespace from simple folder names if no additional information is
    # provided:

    valid_roots = [Path("uavcan"), Path("cyphal")]

    # absolute dsdl path using valid roots
    root = DSDLDefinition._infer_path_to_root_from_first_found(
        Path("/repo/uavcan/foo/bar/435.baz.1.0.dsdl").resolve(), valid_roots
    )
    assert root == Path("/repo/uavcan").resolve()

    # relative dsdl path using valid roots
    root = DSDLDefinition._infer_path_to_root_from_first_found(
        Path("repo/uavcan/foo/bar/435.baz.1.0.dsdl"), valid_roots
    )
    assert root == Path("repo/uavcan")

    # absolute dsdl path using valid roots but an invalid file path
    with expect_raises(PathInferenceError):
        _ = DSDLDefinition._infer_path_to_root_from_first_found(
            Path("/repo/crap/foo/bar/435.baz.1.0.dsdl").resolve(), valid_roots
        )

    # relative dsdl path using valid roots but an invalid file path
    with expect_raises(PathInferenceError):
        _ = DSDLDefinition._infer_path_to_root_from_first_found(Path("repo/crap/foo/bar/435.baz.1.0.dsdl"), valid_roots)

    # relative dsdl path with invalid root fragments
    invalid_root_fragments = [Path("cyphal", "acme")]
    with expect_raises(PathInferenceError):
        _ = DSDLDefinition._infer_path_to_root_from_first_found(
            Path("repo/crap/foo/bar/435.baz.1.0.dsdl"), invalid_root_fragments
        )

    # In this example, foo/bar might look like a valid root path but it is not relative to repo/uavcan/foo/bar and is
    # not considered after relative path inference has failed because it is not a simple root name.
    root = DSDLDefinition._infer_path_to_root_from_first_found(
        Path("repo/uavcan/foo/bar/435.baz.1.0.dsdl"), [Path("foo/bar"), Path("foo")]
    )
    assert root == Path("repo/uavcan/foo")

    # when foo/bar is placed within the proper, relative path it is considered as a valid root and is preferred over
    # the simple root name "foo":
    root = DSDLDefinition._infer_path_to_root_from_first_found(
        Path("repo/uavcan/foo/bar/435.baz.1.0.dsdl"), [Path("repo/uavcan/foo/bar"), Path("foo")]
    )
    assert root == Path("repo/uavcan/foo/bar")

    # Sometimes the root paths have crap in them and need to be resolved:

    root = DSDLDefinition._infer_path_to_root_from_first_found(
        Path("/path/to/repo/uavcan/foo/bar/435.baz.1.0.dsdl").resolve(),
        [Path("/path/to/repo/uavcan/../uavcan").resolve()],
    )
    assert root == Path("/path/to/repo/uavcan").resolve()

    # Let's ensure ordering here

    root = DSDLDefinition._infer_path_to_root_from_first_found(
        Path("repo/uavcan/foo/bar/435.baz.1.0.dsdl"), [Path("repo/uavcan"), Path("repo/uavcan/foo")]
    )
    assert root == Path("repo/uavcan")


def _unittest_type_from_path_inference_edge_case(temp_dsdl_factory) -> None:  # type: ignore
    """
    Edge case where we target a file where the namespace is under the root path.
    """
    # pylint: disable=protected-access

    from pytest import raises as expect_raises
    import os

    target_path = Path("dsdl_root/Type.1.0.dsdl")
    target_file = temp_dsdl_factory.new_file(target_path, "@sealed").resolve()
    expected_root_parent = target_file.parent.parent
    with expect_raises(PathInferenceError):
        _ = DSDLDefinition._infer_path_to_root_from_first_found(target_file, [])

    old_cwd = os.getcwd()
    os.chdir(expected_root_parent)
    try:
        root = DSDLDefinition._infer_path_to_root_from_first_found(target_path, [])
        assert root.parent.resolve() == expected_root_parent
    finally:
        os.chdir(old_cwd)


def _unittest_from_first_in(temp_dsdl_factory) -> None:  # type: ignore
    dsdl_file = temp_dsdl_factory.new_file(Path("repo/uavcan/foo/bar/435.baz.1.0.dsdl"), "@sealed")
    dsdl_def = DSDLDefinition.from_first_in(dsdl_file.resolve(), [(dsdl_file.parent.parent / "..")])
    assert dsdl_def.full_name == "uavcan.foo.bar.baz"



# Copyright (c) 2018 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import logging
from pathlib import Path
from typing import Callable, Iterable

from . import _data_schema_builder, _error, _expression, _parser, _port_id_ranges, _serializable
from ._dsdl import DefinitionVisitor, ReadableDSDLFile


class AssertionCheckFailureError(_error.InvalidDefinitionError):
    pass


class UndefinedDataTypeError(_error.InvalidDefinitionError):
    pass


class UndefinedIdentifierError(_error.InvalidDefinitionError):
    pass


class InvalidDirectiveError(_error.InvalidDefinitionError):
    pass


class UnregulatedFixedPortIDError(_error.InvalidDefinitionError):
    pass


class MissingSerializationModeError(_error.InvalidDefinitionError):
    pass


class DataTypeCollisionError(_error.InvalidDefinitionError):
    """
    Raised when there are conflicting data type definitions.
    """


class DataTypeNameCollisionError(DataTypeCollisionError):
    """
    Raised when type collisions are caused by naming conflicts.
    """


_logger = logging.getLogger(__name__)


class DataTypeBuilder(_parser.StatementStreamProcessor):

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        definition: ReadableDSDLFile,
        lookup_definitions: Iterable[ReadableDSDLFile],
        definition_visitors: Iterable[DefinitionVisitor],
        print_output_handler: Callable[[int, str], None],
        allow_unregulated_fixed_port_id: bool,
    ):
        self._definition = definition
        self._lookup_definitions = list(lookup_definitions)
        self._definition_visitors = definition_visitors
        self._print_output_handler = print_output_handler
        self._allow_unregulated_fixed_port_id = allow_unregulated_fixed_port_id
        self._element_callback = None  # type: Callable[[str], None] | None

        assert isinstance(self._definition, ReadableDSDLFile)
        assert all(map(lambda x: isinstance(x, ReadableDSDLFile), lookup_definitions))
        assert callable(self._print_output_handler)
        assert isinstance(self._allow_unregulated_fixed_port_id, bool)

        self._structs = [_data_schema_builder.DataSchemaBuilder()]
        self._is_deprecated = False

    def finalize(self) -> _serializable.CompositeType:
        if len(self._structs) == 1:  # Structure type
            (builder,) = self._structs
            out = self._make_composite(
                builder=builder,
                name=self._definition.full_name,
                version=self._definition.version,
                deprecated=self._is_deprecated,
                fixed_port_id=self._definition.fixed_port_id,
                source_file_path=self._definition.file_path,
                has_parent_service=False,
            )
        else:  # Service type
            request_builder, response_builder = self._structs
            assert isinstance(request_builder, _data_schema_builder.DataSchemaBuilder)
            assert isinstance(response_builder, _data_schema_builder.DataSchemaBuilder)
            sep = _serializable.CompositeType.NAME_COMPONENT_SEPARATOR
            request = self._make_composite(
                builder=request_builder,
                name=sep.join([self._definition.full_name, "Request"]),
                version=self._definition.version,
                deprecated=self._is_deprecated,
                fixed_port_id=None,
                source_file_path=self._definition.file_path,
                has_parent_service=True,
            )
            response = self._make_composite(
                builder=response_builder,
                name=sep.join([self._definition.full_name, "Response"]),
                version=self._definition.version,
                deprecated=self._is_deprecated,
                fixed_port_id=None,
                source_file_path=self._definition.file_path,
                has_parent_service=True,
            )
            # noinspection SpellCheckingInspection
            out = _serializable.ServiceType(  # pozabito vse na svete
                request=request,  # serdce zamerlo v grudi
                response=response,  # tolko nebo tolko veter
                fixed_port_id=self._definition.fixed_port_id,  # tolko radost vperedi
            )

        assert isinstance(out, _serializable.CompositeType)
        if not self._allow_unregulated_fixed_port_id:
            port_id = out.fixed_port_id
            if port_id is not None:
                is_service_type = isinstance(out, _serializable.ServiceType)
                f = (
                    _port_id_ranges.is_valid_regulated_service_id
                    if is_service_type
                    else _port_id_ranges.is_valid_regulated_subject_id
                )
                if not f(port_id, out.root_namespace):
                    raise UnregulatedFixedPortIDError(
                        "Regulated port ID %r for %s type %r is not valid. "
                        "Consider using allow_unregulated_fixed_port_id."
                        % (port_id, "service" if is_service_type else "message", out.full_name)
                    )
        return out

    def on_attribute_comment(self, comment: str) -> None:
        # Flush queued attribute with doc comment
        self._flush_attribute(comment)

    def on_header_comment(self, comment: str) -> None:
        # Attach doc to composite type
        self._structs[-1].set_comment(comment)

    def on_constant(self, constant_type: _serializable.SerializableType, name: str, value: _expression.Any) -> None:
        self._on_attribute()
        self._queue_attribute(
            lambda doc: self._structs[-1].add_constant(_serializable.Constant(constant_type, name, value, doc))
        )

    def on_field(self, field_type: _serializable.SerializableType, name: str) -> None:
        self._on_attribute()
        self._queue_attribute(lambda doc: self._structs[-1].add_field(_serializable.Field(field_type, name, doc)))

    def on_padding_field(self, padding_field_type: _serializable.VoidType) -> None:
        self._on_attribute()
        self._queue_attribute(
            lambda doc: self._structs[-1].add_field(_serializable.PaddingField(padding_field_type, doc))
        )

    def on_directive(
        self, line_number: int, directive_name: str, associated_expression_value: _expression.Any | None
    ) -> None:
        try:
            handler = {
                "print": self._on_print_directive,
                "assert": self._on_assert_directive,
                "extent": self._on_extent_directive,
                "sealed": self._on_sealed_directive,
                "union": self._on_union_directive,
                "deprecated": self._on_deprecated_directive,
            }[directive_name]
        except KeyError:
            raise InvalidDirectiveError("Unknown directive %r" % directive_name) from None
        else:
            assert callable(handler)
            return handler(line_number, associated_expression_value)

    def on_service_response_marker(self) -> None:
        if len(self._structs) > 1:
            raise _error.InvalidDefinitionError("Duplicated service response marker")

        self._structs.append(_data_schema_builder.DataSchemaBuilder())
        assert len(self._structs) == 2

    def resolve_top_level_identifier(self, name: str) -> _expression.Any:
        # Look only in the current data structure. The lookup cannot cross the service request/response boundary.
        for c in self._structs[-1].constants:
            if c.name == name:
                return c.value

        if name == "_offset_":
            bls = self._structs[-1].offset
            assert len(bls) > 0 and all(map(lambda x: isinstance(x, int), bls))
            # FIXME: THIS OPERATION TRIGGERS NUMERICAL EXPANSION OF THE BIT LENGTH SET.
            # TODO: INTEGRATE THE SET EXPRESSION WITH THE BIT LENGTH SET SOLVER TO IMPROVE PERFORMANCE.
            return _expression.Set(map(_expression.Rational, bls))
        raise UndefinedIdentifierError("Undefined identifier: %r" % name)

    def resolve_versioned_data_type(self, name: str, version: _serializable.Version) -> _serializable.CompositeType:
        if _serializable.CompositeType.NAME_COMPONENT_SEPARATOR in name:
            full_name = name
        else:
            full_name = _serializable.CompositeType.NAME_COMPONENT_SEPARATOR.join(
                [self._definition.full_namespace, name]
            )
            _logger.debug("The full name of a relatively referred type %r reconstructed as %r", name, full_name)

        del name
        found = list(
            filter(
                lambda d: d.full_name.lower() == full_name.lower() and d.version == version, self._lookup_definitions
            )
        )
        if not found:
            # Play Sherlock to help the user with mistakes like https://forum.opencyphal.org/t/904/2
            requested_ns = full_name.split(_serializable.CompositeType.NAME_COMPONENT_SEPARATOR)[0]
            lookup_nss = set(x.root_namespace for x in self._lookup_definitions)
            subroot_ns = self._definition.name_components[1] if len(self._definition.name_components) > 2 else None
            error_description = "Data type %s.%d.%d could not be found in the following root namespaces: %s. " % (
                full_name,
                version.major,
                version.minor,
                lookup_nss or "(empty set)",
            )
            if requested_ns not in lookup_nss and requested_ns == subroot_ns:
                error_description += " Did you mean to use the directory %s instead of %s?" % (
                    self._definition.root_namespace_path / subroot_ns,
                    self._definition.root_namespace_path,
                )
            else:
                error_description += " Please make sure that you specified the directories correctly."
            raise UndefinedDataTypeError(error_description)

        if len(found) > 1:
            if (
                found[0].full_name != found[1].full_name and found[0].full_name.lower() == found[1].full_name.lower()
            ):  # pragma: no cover
                # This only happens if the file system is case-insensitive.
                raise DataTypeNameCollisionError(
                    "Full name of this definition differs from %s only by letter case, "
                    "which is not permitted" % found[0].file_path,
                    path=found[1].file_path,
                )
            raise DataTypeCollisionError("Conflicting definitions: %r" % found)
        elif found[0].full_name != full_name and found[0].full_name.lower() == full_name.lower():
            # pragma: no cover
            # This only happens if the file system is case-sensitive.
            raise DataTypeNameCollisionError(
                "Full name of required definition %s differs from %s only by letter case, "
                "which is not permitted" % (full_name, found[0].full_name),
                path=found[0].file_path,
            )

        target_definition = found[0]

        assert isinstance(target_definition, ReadableDSDLFile)
        assert target_definition.version == version

        for visitor in self._definition_visitors:
            visitor.on_definition(self._definition, target_definition)

        # Recursion is cool.
        dt = target_definition.read(
            lookup_definitions=self._lookup_definitions,
            definition_visitors=self._definition_visitors,
            print_output_handler=self._print_output_handler,
            allow_unregulated_fixed_port_id=self._allow_unregulated_fixed_port_id,
        )
        return dt

    def _queue_attribute(self, element_callback: Callable[[str], None]) -> None:
        self._flush_attribute("")
        self._element_callback = element_callback

    def _flush_attribute(self, comment: str) -> None:
        if self._element_callback is not None:
            self._element_callback(comment)
        self._element_callback = None

    def _on_attribute(self) -> None:
        if isinstance(self._structs[-1].serialization_mode, _data_schema_builder.DelimitedSerializationMode):
            raise InvalidDirectiveError(
                "The extent directive can only be placed after the last attribute definition in the schema. "
                "This is to prevent errors if the extent is dependent on the bit length set of the data schema."
            )

    def _on_print_directive(self, line_number: int, value: _expression.Any | None) -> None:
        _logger.info(
            "Print directive at %s:%d%s",
            self._definition.file_path,
            line_number,
            (": %s" % value) if value is not None else " (no value to print)",
        )
        self._print_output_handler(line_number, str(value if value is not None else ""))

    def _on_assert_directive(self, line_number: int, value: _expression.Any | None) -> None:
        if isinstance(value, _expression.Boolean):
            if not value.native_value:
                raise AssertionCheckFailureError(
                    "Assertion check has failed", path=self._definition.file_path, line=line_number
                )
            _logger.debug("Assertion check successful at %s:%d", self._definition.file_path, line_number)
        elif value is None:
            raise InvalidDirectiveError("Assert directive requires an expression")
        else:
            raise InvalidDirectiveError("The assertion check expression must yield a boolean, not %s" % value.TYPE_NAME)

    def _on_extent_directive(self, line_number: int, value: _expression.Any | None) -> None:
        if self._structs[-1].serialization_mode is not None:
            raise InvalidDirectiveError(
                "Misplaced extent directive. The serialization mode is already set to %s"
                % self._structs[-1].serialization_mode
            )
        if value is None:
            raise InvalidDirectiveError("The extent directive requires an expression")
        if isinstance(value, _expression.Rational):
            struct = self._structs[-1]
            bits = value.as_native_integer()
            struct.set_serialization_mode(_data_schema_builder.DelimitedSerializationMode(bits))
            _logger.debug("The extent is set to %d bits at %s:%d", bits, self._definition.file_path, line_number)
        else:
            raise InvalidDirectiveError("The extent directive expects a rational, not %s" % value.TYPE_NAME)

    def _on_sealed_directive(self, _ln: int, value: _expression.Any | None) -> None:
        if self._structs[-1].serialization_mode is not None:
            raise InvalidDirectiveError(
                "Misplaced sealing directive. The serialization mode is already set to %s"
                % self._structs[-1].serialization_mode
            )
        if value is not None:
            raise InvalidDirectiveError("The sealed directive does not expect an expression")
        self._structs[-1].set_serialization_mode(_data_schema_builder.SealedSerializationMode())

    def _on_union_directive(self, _ln: int, value: _expression.Any | None) -> None:
        if value is not None:
            raise InvalidDirectiveError("The union directive does not expect an expression")
        if self._structs[-1].union:
            raise InvalidDirectiveError("Duplicated union directive")
        if self._structs[-1].attributes:
            raise InvalidDirectiveError("The union directive must be placed before the first " "attribute definition")
        self._structs[-1].make_union()

    def _on_deprecated_directive(self, _ln: int, value: _expression.Any | None) -> None:
        if value is not None:
            raise InvalidDirectiveError("The deprecated directive does not expect an expression")
        if self._is_deprecated:
            raise InvalidDirectiveError("Duplicated deprecated directive")
        if len(self._structs) > 1:
            raise InvalidDirectiveError("The deprecated directive cannot be placed in the response section")
        if self._structs[-1].attributes:
            raise InvalidDirectiveError(
                "The deprecated directive must be placed before the first " "attribute definition"
            )
        self._is_deprecated = True

    @staticmethod
    def _make_composite(  # pylint: disable=too-many-arguments
        builder: _data_schema_builder.DataSchemaBuilder,
        name: str,
        version: _serializable.Version,
        deprecated: bool,
        fixed_port_id: int | None,
        source_file_path: Path,
        has_parent_service: bool,
    ) -> _serializable.CompositeType:
        ty = _serializable.UnionType if builder.union else _serializable.StructureType
        inner = ty(
            name=name,
            version=version,
            attributes=builder.attributes,
            deprecated=deprecated,
            fixed_port_id=fixed_port_id,
            source_file_path=source_file_path,
            has_parent_service=has_parent_service,
            doc=builder.doc,
        )  # type: _serializable.CompositeType
        sm = builder.serialization_mode
        if isinstance(sm, _data_schema_builder.DelimitedSerializationMode):
            out = _serializable.DelimitedType(inner, extent=sm.extent)  # type: _serializable.CompositeType
            _logger.debug("%r wrapped into %r", inner, out)
        elif isinstance(sm, _data_schema_builder.SealedSerializationMode):
            out = inner
        else:
            assert sm is None, "I wish Python had a strong static type system"
            raise MissingSerializationModeError(
                "%s: Either `@sealed` or `@extent ...` are required. "
                "The smallest valid extent for this type (i.e., its max bit length) is %d bits (%d bytes). "
                "If you are not sure what this means, add the following line near the end of this definition: "
                "`@extent %d * 8`"
                % (inner.short_name, inner.extent, inner.extent // 8, DataTypeBuilder._suggest_extent_in_bytes(inner))
            )

        return out

    @staticmethod
    def _suggest_extent_in_bytes(model: _serializable.CompositeType) -> int:
        """
        An implementation-specific heuristic intended to lower the entry barrier.
        The numbers are subject to change between minor revisions.
        """
        return max(64, model.extent * 2 // 8)



# Copyright (c) 2021 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

import random
import itertools
from ._symbolic import NullaryOperator, validate_numerically


def _unittest_nullary() -> None:
    import pytest

    op = NullaryOperator([0])
    assert set(op.expand()) == {0}
    assert set(op.modulo(12345)) == {0}
    assert op.min == op.max == 0
    validate_numerically(op)

    op = NullaryOperator([1, 2, 3, 4, 5, 6, 7, 8])
    assert set(op.expand()) == {1, 2, 3, 4, 5, 6, 7, 8}
    assert set(op.modulo(4)) == {0, 1, 2, 3}
    assert (op.min, op.max) == (1, 8)
    validate_numerically(op)

    with pytest.raises(ValueError):
        NullaryOperator([])


def _unittest_padding() -> None:
    from ._symbolic import PaddingOperator

    op = PaddingOperator(
        NullaryOperator([1, 2, 3, 4, 5, 6, 7, 8, 9]),
        4,
    )
    assert op.min == 4
    assert op.max == 12
    assert set(op.expand()) == {4, 8, 12}
    assert set(op.modulo(2)) == {0}
    assert set(op.modulo(4)) == {0}
    assert set(op.modulo(8)) == {0, 4}
    assert set(op.modulo(16)) == {4, 8, 12}
    validate_numerically(op)

    assert set(x % 6 for x in op.expand()) == {0, 2, 4}  # Reference
    assert set(op.modulo(6)) == {0, 2, 4}

    assert set(x % 7 for x in op.expand()) == {1, 4, 5}  # Reference
    assert set(op.modulo(7)) == {1, 4, 5}

    for _ in range(1):
        child = NullaryOperator(random.randint(0, 1024) for _ in range(random.randint(1, 100)))
        alignment = random.randint(1, 64)
        op = PaddingOperator(child, alignment)
        div = random.randint(1, 64)
        assert set(op.modulo(div)) == {x % div for x in op.expand()}
        validate_numerically(op)


def _unittest_concatenation() -> None:
    import pytest
    from ._symbolic import ConcatenationOperator

    op = ConcatenationOperator(
        [
            NullaryOperator([1]),
            NullaryOperator([2]),
            NullaryOperator([10]),
        ]
    )
    assert op.min == op.max == 13
    assert set(op.expand()) == {13}
    assert set(op.modulo(1)) == {0}
    assert set(op.modulo(2)) == {1}
    assert set(op.modulo(13)) == {0}
    assert set(op.modulo(8)) == {5}
    validate_numerically(op)

    op = ConcatenationOperator(
        [
            NullaryOperator([1, 2, 10]),
        ]
    )
    assert op.min == 1
    assert op.max == 10
    assert set(op.expand()) == {1, 2, 10}
    assert set(op.modulo(1)) == {0}
    assert set(op.modulo(2)) == {0, 1}
    assert set(op.modulo(8)) == {1, 2}
    validate_numerically(op)

    op = ConcatenationOperator(
        [
            NullaryOperator([1, 2]),
            NullaryOperator([4, 5]),
        ]
    )
    assert op.min == 5
    assert op.max == 7
    assert set(op.expand()) == {5, 6, 7}
    assert set(op.modulo(5)) == {0, 1, 2}
    assert set(op.modulo(8)) == {5, 6, 7}
    validate_numerically(op)

    op = ConcatenationOperator(
        [
            NullaryOperator([1, 2, 3]),
            NullaryOperator([4, 5, 6]),
            NullaryOperator([7, 8, 9]),
        ]
    )
    assert op.min == 12
    assert op.max == 18
    assert set(op.expand()) == {12, 13, 14, 15, 16, 17, 18}
    assert set(op.modulo(8)) == {0, 1, 2, 4, 5, 6, 7}  # 3 is missing
    validate_numerically(op)

    for _ in range(1):
        op = ConcatenationOperator(
            [
                NullaryOperator(random.randint(0, 1024) for _ in range(random.randint(1, 10)))
                for _ in range(random.randint(1, 10))
            ]
        )
        div = random.randint(1, 64)
        assert set(op.modulo(div)) == {x % div for x in op.expand()}
        validate_numerically(op)

    with pytest.raises(ValueError):
        ConcatenationOperator([])


def _unittest_repetition() -> None:
    from ._symbolic import RepetitionOperator

    op = RepetitionOperator(
        NullaryOperator([7, 11, 17]),
        3,
    )
    assert op.min == 7 * 3
    assert op.max == 17 * 3
    assert set(op.expand()) == set(map(sum, itertools.combinations_with_replacement([7, 11, 17], 3)))
    assert set(op.expand()) == {21, 25, 29, 31, 33, 35, 39, 41, 45, 51}
    assert set(op.modulo(7)) == {0, 1, 2, 3, 4, 5, 6}
    assert set(op.modulo(8)) == {1, 3, 5, 7}
    validate_numerically(op)

    for _ in range(1):
        child = NullaryOperator(random.randint(0, 100) for _ in range(random.randint(1, 10)))
        k = random.randint(0, 10)
        ref = set(map(sum, itertools.combinations_with_replacement(child.expand(), k)))
        op = RepetitionOperator(child, k)
        assert set(op.expand()) == ref

        assert op.min == min(child.expand()) * k
        assert op.max == max(child.expand()) * k

        div = random.randint(1, 64)
        assert set(op.modulo(div)) == {x % div for x in ref}

        validate_numerically(op)


def _unittest_range_repetition() -> None:
    from ._symbolic import RangeRepetitionOperator

    op = RangeRepetitionOperator(
        NullaryOperator([7, 11, 17]),
        3,
    )
    assert op.min == 0  # Always 0
    assert op.max == 17 * 3
    assert set(op.expand()) == (
        {0}
        | set(map(sum, itertools.combinations_with_replacement([7, 11, 17], 1)))
        | set(map(sum, itertools.combinations_with_replacement([7, 11, 17], 2)))
        | set(map(sum, itertools.combinations_with_replacement([7, 11, 17], 3)))
    )
    assert set(op.expand()) == {0, 7, 11, 14, 17, 18, 21, 22, 24, 25, 28, 29, 31, 33, 34, 35, 39, 41, 45, 51}
    assert set(op.modulo(7)) == {0, 1, 2, 3, 4, 5, 6}
    validate_numerically(op)

    op = RangeRepetitionOperator(
        NullaryOperator([7, 11]),
        2,
    )
    assert op.min == 0  # Always 0
    assert op.max == 22
    assert set(op.expand()) == {0, 7, 14, 11, 18, 22}
    assert set(op.modulo(7)) == {0, 1, 4}
    assert set(op.modulo(8)) == {0, 2, 3, 6, 7}
    validate_numerically(op)

    for _ in range(1):
        child = NullaryOperator(random.randint(0, 100) for _ in range(random.randint(1, 10)))
        k_max = random.randint(0, 10)
        ref = set(
            itertools.chain(
                *(map(sum, itertools.combinations_with_replacement(child.expand(), k)) for k in range(k_max + 1))
            )
        )
        op = RangeRepetitionOperator(child, k_max)
        assert set(op.expand()) == ref

        assert op.min == 0
        assert op.max == max(child.expand()) * k_max

        div = random.randint(1, 64)
        assert set(op.modulo(div)) == {x % div for x in ref}

        validate_numerically(op)


def _unittest_union() -> None:
    import pytest
    from ._symbolic import UnionOperator

    op = UnionOperator(
        [
            NullaryOperator([1, 2, 3]),
            NullaryOperator([4, 5, 6]),
            NullaryOperator([7, 8, 9]),
        ]
    )
    assert op.min == 1
    assert op.max == 9
    assert set(op.expand()) == {1, 2, 3, 4, 5, 6, 7, 8, 9}
    assert set(op.modulo(8)) == {x % 8 for x in op.expand()}
    validate_numerically(op)

    op = UnionOperator(
        [
            NullaryOperator([13, 17, 21, 29]),
            NullaryOperator([8, 16]),
        ]
    )
    assert op.min == 8
    assert op.max == 29
    assert set(op.expand()) == {13, 17, 21, 29, 8, 16}
    assert set(op.modulo(7)) == {x % 7 for x in op.expand()}
    assert set(op.modulo(8)) == {x % 8 for x in op.expand()}
    validate_numerically(op)

    for _ in range(1):
        op = UnionOperator(
            [
                NullaryOperator(random.randint(0, 1024) for _ in range(random.randint(1, 10)))
                for _ in range(random.randint(1, 10))
            ]
        )
        validate_numerically(op)

    with pytest.raises(ValueError):
        UnionOperator([])


def _unittest_repr() -> None:
    from ._symbolic import (
        PaddingOperator,
        ConcatenationOperator,
        RepetitionOperator,
        RangeRepetitionOperator,
        UnionOperator,
        MemoizationOperator,
    )

    op = MemoizationOperator(
        UnionOperator(
            [
                PaddingOperator(NullaryOperator([1, 2, 3, 4, 5, 6, 7, 8]), 4),
                ConcatenationOperator(
                    [
                        NullaryOperator([8, 16]),
                        NullaryOperator([96, 112, 120]),
                        RangeRepetitionOperator(NullaryOperator([64]), 8),
                    ]
                ),
                RepetitionOperator(
                    UnionOperator(
                        [
                            NullaryOperator([32]),
                            NullaryOperator([40]),
                        ]
                    ),
                    2,
                ),
            ]
        )
    )
    validate_numerically(op)
    assert repr(op) == "(pad(4,{1,2,3,4,5,6,7,8})|concat({8,16},{96,112,120},repeat(<=8,{64}))|repeat(2,({32}|{40})))"



# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT

from __future__ import annotations
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Callable, Iterable, List, Tuple, TypeVar

from ._serializable import CompositeType, Version

PrintOutputHandler = Callable[[Path, int, str], None]
"""Invoked when the frontend encounters a print directive or needs to output a generic diagnostic."""


class DSDLFile(ABC):
    """
    Interface for DSDL files. This interface is used by the parser to abstract DSDL type details inferred from the
    filesystem. Where properties are duplicated between the composite type and this file the composite type is to be
    considered canonical. The properties directly on this class are inferred from the dsdl file path before the
    composite type has been parsed.
    """

    @property
    @abstractmethod
    def composite_type(self) -> CompositeType | None:
        """The composite type that was read from the DSDL file or None if the type has not been parsed yet."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def full_name(self) -> str:
        """The full name, e.g., uavcan.node.Heartbeat"""
        raise NotImplementedError()

    @property
    def name_components(self) -> List[str]:
        """Components of the full name as a list, e.g., ['uavcan', 'node', 'Heartbeat']"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def short_name(self) -> str:
        """The last component of the full name, e.g., Heartbeat of uavcan.node.Heartbeat"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def full_namespace(self) -> str:
        """The full name without the short name, e.g., uavcan.node for uavcan.node.Heartbeat"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def root_namespace(self) -> str:
        """The first component of the full name, e.g., uavcan of uavcan.node.Heartbeat"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def text(self) -> str:
        """The source text in its raw unprocessed form (with comments, formatting intact, and everything)"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def version(self) -> Version:
        """
        The version of the DSDL definition.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def fixed_port_id(self) -> int | None:
        """Either the fixed port ID as integer, or None if not defined for this type."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def has_fixed_port_id(self) -> bool:
        """
        If the type has a fixed port ID defined, this method returns True. Equivalent to ``fixed_port_id is not None``.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def file_path(self) -> Path:
        """The path to the DSDL file on the filesystem."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def root_namespace_path(self) -> Path:
        """
        The path to the root namespace directory on the filesystem.
        """
        raise NotImplementedError()


class ReadableDSDLFile(DSDLFile):
    """
    A DSDL file that can construct a composite type from its contents.
    """

    @property
    @abstractmethod
    def file_path(self) -> Path:
        """
        Path to an existing DSDL file on the filesystem.
        """
        raise NotImplementedError()

    @abstractmethod
    def read(
        self,
        lookup_definitions: Iterable["ReadableDSDLFile"],
        definition_visitors: Iterable["DefinitionVisitor"],
        print_output_handler: Callable[[int, str], None],
        allow_unregulated_fixed_port_id: bool,
    ) -> CompositeType:
        """
        Reads the data type definition and returns its high-level data type representation.
        The output should be cached; all following invocations should read from this cache.
        Caching is very important, because it is expected that the same definition may be referred to multiple
        times (e.g., for composition or when accessing external constants). Re-processing a definition every time
        it is accessed would be a huge waste of time.
        Note, however, that this may lead to unexpected complications if one is attempting to re-read a definition
        with different inputs (e.g., different lookup paths) expecting to get a different result: caching would
        get in the way. That issue is easy to avoid by creating a new instance of the object.
        :param lookup_definitions:              List of definitions available for referring to.
        :param definition_visitors:             Visitors to notify about discovered dependencies.
        :param print_output_handler:            Used for @print and for diagnostics: (line_number, text) -> None.
        :param allow_unregulated_fixed_port_id: Do not complain about fixed unregulated port IDs.
        :return: The data type representation.
        """
        raise NotImplementedError()


class DefinitionVisitor(ABC):
    """
    A visitor that is notified about discovered dependencies.
    """

    @abstractmethod
    def on_definition(self, target_dsdl_file: DSDLFile, dependency_dsdl_file: ReadableDSDLFile) -> None:
        """
        Called by the parser after if finds a dependent type but before it parses a file in a lookup namespace.
        :param target_dsdl_file: The target DSDL file that has dependencies the parser is searching for.
        :param dependency_dsdl_file: The dependency of target_dsdl_file file the parser is about to parse.
        """
        raise NotImplementedError()


SortedFileT = TypeVar("SortedFileT", DSDLFile, ReadableDSDLFile, CompositeType)
SortedFileList = List[SortedFileT]
"""A list of DSDL files sorted by name, newest version first."""


def get_definition_ordering_rank(d: DSDLFile | CompositeType) -> Tuple[str, int, int]:
    return d.full_name, -d.version.major, -d.version.minor


def file_sort(file_list: Iterable[SortedFileT]) -> SortedFileList[SortedFileT]:
    """
    Sorts a list of DSDL files lexicographically by name, newest version first.
    """
    return list(sorted(file_list, key=get_definition_ordering_rank))


def normalize_paths_argument_to_list(namespaces_or_namespace: None | Path | str | Iterable[Path | str]) -> List[Path]:
    """
    Normalizes the input argument to a list of paths.
    """
    if namespaces_or_namespace is None:
        return []
    if isinstance(namespaces_or_namespace, (Path, str)):
        return [Path(namespaces_or_namespace)]

    def _convert(arg: Any) -> Path:
        if not isinstance(arg, (str, Path)):
            raise TypeError(f"Invalid type: {type(arg)}")
        return Path(arg) if isinstance(arg, str) else arg

    value_set = set()

    def _filter_duplicate_paths(arg: Any) -> bool:
        if arg in value_set:
            return False
        value_set.add(arg)
        return True

    converted = [_convert(arg) for arg in namespaces_or_namespace]
    return list(filter(_filter_duplicate_paths, converted))


# +-[UNIT TESTS]------------------------------------------------------------------------------------------------------+


def _unittest_dsdl_normalize_paths_argument_to_list() -> None:

    from pytest import raises as assert_raises

    # Test with None argument
    result = normalize_paths_argument_to_list(None)
    assert result == []

    # Test with single string argument
    result = normalize_paths_argument_to_list("path/to/namespace")
    assert result == [Path("path/to/namespace")]

    # Test with single Path argument
    result = normalize_paths_argument_to_list(Path("path/to/namespace"))
    assert result == [Path("path/to/namespace")]

    # Test with list of strings argument
    result = normalize_paths_argument_to_list(["path/to/namespace1", "path/to/namespace2"])
    assert result == [Path("path/to/namespace1"), Path("path/to/namespace2")]

    # Test with list of Path arguments
    result = normalize_paths_argument_to_list([Path("path/to/namespace1"), Path("path/to/namespace2")])
    assert result == [Path("path/to/namespace1"), Path("path/to/namespace2")]

    # Test with mixed list of strings and Path arguments
    result = normalize_paths_argument_to_list(["path/to/namespace1", Path("path/to/namespace2")])
    assert result == [Path("path/to/namespace1"), Path("path/to/namespace2")]

    # Test de-duplication
    result = normalize_paths_argument_to_list(["path/to/namespace1", "path/to/namespace1"])
    assert result == [Path("path/to/namespace1")]

    # Test with invalid argument type
    with assert_raises(TypeError):
        normalize_paths_argument_to_list(42)  # type: ignore

    # Test with invalid argument type
    with assert_raises(TypeError):
        normalize_paths_argument_to_list([42])  # type: ignore



# Copyright (c) 2018 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

# pylint: disable=logging-not-lazy

from __future__ import annotations
import collections
import itertools
import logging
from itertools import product, repeat
from pathlib import Path
from typing import Callable, DefaultDict, Iterable

from . import _dsdl_definition, _error, _serializable
from ._dsdl import ReadableDSDLFile, PrintOutputHandler, SortedFileList
from ._dsdl import file_sort as dsdl_file_sort
from ._dsdl import normalize_paths_argument_to_list
from ._namespace_reader import DSDLDefinitions, read_definitions

_logger = logging.getLogger(__name__)


class RootNamespaceNameCollisionError(_error.InvalidDefinitionError):
    """
    Raised when there is more than one namespace under the same name.
    This may occur if there are identically named namespaces located in different directories.
    """


class NestedRootNamespaceError(_error.InvalidDefinitionError):
    """
    Nested root namespaces are not allowed. This exception is thrown when this rule is violated.
    """


class FixedPortIDCollisionError(_error.InvalidDefinitionError):
    """
    Raised when there is more than one data type, or different major versions of the same data type
    using the same fixed port ID.
    """


class VersionsOfDifferentKindError(_error.InvalidDefinitionError):
    """
    Definitions that share the same name but are of different kinds.
    """


class MinorVersionFixedPortIDError(_error.InvalidDefinitionError):
    """
    Different fixed port-ID under the same major version, or a fixed port ID was removed under the same major version.
    """


class ExtentConsistencyError(_error.InvalidDefinitionError):
    """
    Different extent under the same major version.
    """


class SealingConsistencyError(_error.InvalidDefinitionError):
    """
    Different sealing status under the same major version.
    """


# +--[PUBLIC API]-----------------------------------------------------------------------------------------------------+


def read_namespace(
    root_namespace_directory: Path | str,
    lookup_directories: None | Path | str | Iterable[Path | str] = None,
    print_output_handler: PrintOutputHandler | None = None,
    allow_unregulated_fixed_port_id: bool = False,
    allow_root_namespace_name_collision: bool = True,
) -> list[_serializable.CompositeType]:
    """
    This function is a main entry point for the library.
    It reads all DSDL definitions from the specified root namespace directory and produces the annotated AST.

    :param root_namespace_directory: The path of the root namespace directory that will be read.
        For example, ``dsdl/uavcan`` to read the ``uavcan`` namespace.

    :param lookup_directories: List of other namespace directories containing data type definitions that are
        referred to from the target root namespace. For example, if you are reading a vendor-specific namespace,
        the list of lookup directories should always include a path to the standard root namespace ``uavcan``,
        otherwise the types defined in the vendor-specific namespace won't be able to use data types from the
        standard namespace.

    :param print_output_handler: If provided, this callable will be invoked when a ``@print`` directive
        is encountered or when the frontend needs to emit a diagnostic;
        the arguments are: path, line number (1-based), text.
        If not provided, no output will be produced except for the standard Python logging subsystem
        (but ``@print`` expressions will be evaluated anyway, and a failed evaluation will be a fatal error).

    :param allow_unregulated_fixed_port_id: Do not reject unregulated fixed port identifiers.
        As demanded by the specification, the frontend rejects unregulated fixed port ID by default.
        This is a dangerous feature that must not be used unless you understand the risks.
        Please read https://opencyphal.org/guide.

    :param allow_root_namespace_name_collision: Allow using the source root namespace name in the look up dirs or
             the same root namespace name multiple times in the lookup dirs. This will enable defining a namespace
             partially and let other entities define new messages or new sub-namespaces in the same root namespace.

    :return: A list of :class:`pydsdl.CompositeType` found under the `root_namespace_directory` and sorted
             lexicographically by full data type name, then by major version (newest version first), then by minor
             version (newest version first). The ordering guarantee allows the caller to always find the newest version
             simply by picking the first matching occurrence.

    :raises: :class:`pydsdl.FrontendError`, :class:`MemoryError`, :class:`SystemError`,
        :class:`OSError` if directories do not exist or inaccessible,
        :class:`ValueError`/:class:`TypeError` if the arguments are invalid.
    """
    # Normalize paths and remove duplicates. Resolve symlinks to avoid ambiguities.
    root_namespace_directory = Path(root_namespace_directory).resolve()

    lookup_directories_path_list = _construct_lookup_directories_path_list(
        [root_namespace_directory],
        normalize_paths_argument_to_list(lookup_directories),
        allow_root_namespace_name_collision,
    )

    # Construct DSDL definitions from the target and the lookup dirs.
    target_dsdl_definitions = _construct_dsdl_definitions_from_namespaces([root_namespace_directory])
    if not target_dsdl_definitions:
        _logger.info("The namespace at %s is empty", root_namespace_directory)
        return []
    _logger.debug("Target DSDL definitions are listed below:")
    for x in target_dsdl_definitions:
        _logger.debug(_LOG_LIST_ITEM_PREFIX + str(x))

    return _complete_read_function(
        target_dsdl_definitions, lookup_directories_path_list, print_output_handler, allow_unregulated_fixed_port_id
    ).direct


# pylint: disable=too-many-arguments
def read_files(
    dsdl_files: None | Path | str | Iterable[Path | str],
    root_namespace_directories_or_names: None | Path | str | Iterable[Path | str],
    lookup_directories: None | Path | str | Iterable[Path | str] = None,
    print_output_handler: PrintOutputHandler | None = None,
    allow_unregulated_fixed_port_id: bool = False,
) -> tuple[list[_serializable.CompositeType], list[_serializable.CompositeType]]:
    """
    This function is a main entry point for the library.
    It reads all DSDL definitions from the specified ``dsdl_files`` and produces the annotated AST for these types and
    the transitive closure of the types they depend on.

    :param Any dsdl_files: A list of paths to dsdl files to parse.

    :param root_namespace_directories_or_names: This can be a set of names of root namespaces or paths to root
        namespaces. All ``dsdl_files`` provided must be under one of these roots. For example, given:

        .. code-block:: python

            dsdl_files = [
                            Path("workspace/project/types/animals/felines/Tabby.1.0.dsdl"),
                            Path("workspace/project/types/animals/canines/Boxer.1.0.dsdl"),
                            Path("workspace/project/types/plants/trees/DouglasFir.1.0.dsdl")
                         ]


        then this argument must be one of:

        .. code-block:: python

            root_namespace_directories_or_names = ["animals", "plants"]
            # or
            root_namespace_directories_or_names = [
                                                    Path("workspace/project/types/animals"),
                                                    Path("workspace/project/types/plants")
                                                  ]
            # or
            root_namespace_directories_or_names = [
                                                    Path("animals"),
                                                    Path("workspace/project/types/plants")
                                                  ]
            # etc


        Any target path not found on the filesystem must be located under a root namespace directory. For example,
        given a set of relative target files like this:

        .. code-block:: python

            dsdl_files = [
                            Path("animals/felines/Tabby.1.0.dsdl"),
                            Path("animals/canines/Boxer.1.0.dsdl"),
                            Path("plants/trees/DouglasFir.1.0.dsdl")
                         ]

        then this argument must include paths under which the target files can be found. For example:

        .. code-block:: python

            root_namespace_directories_or_names = [
                                                    Path("workspace/project/types/animals"),
                                                    Path("workspace/project/types/plants")
                                                  ]

        Note that relative paths shown here would only resolve correctly if the current working directory is set to
        the root of the workspace (this function does not search system paths for the target files). If the target files
        are located in a different folder the paths must be absolute. For example:

        .. code-block:: python

            root_namespace_directories_or_names = [
                                                    Path("/path/to/workspace/project/types/animals"),
                                                    Path("/path/to/workspace/project/types/plants")
                                                  ]

        When reading vendor-specific types, the list of lookup directories should always include a path to the standard
        root namespace ``uavcan``, otherwise the types defined in the vendor-specific namespace won't be able to use
        data types from the standard namespace.

    :param lookup_directories: List of namespace directories containing data type definitions that are referred to from
        the target dsdl files. Callers should prefer combining the lookup directories with the root namespace
        directories unless certain paths should be excluded when looking for target types. That is, only
        ``root_namespace_directories_or_names`` is used to find the target types but both this list and
        ``root_namespace_directories_or_names`` are used to find the dependent types.

    :param print_output_handler: If provided, this callable will be invoked when a ``@print`` directive
        is encountered or when the frontend needs to emit a diagnostic;
        the arguments are: path, line number (1-based), text.
        If not provided, no output will be produced except for the standard Python logging subsystem
        (but ``@print`` expressions will be evaluated anyway, and a failed evaluation will be a fatal error).

    :param allow_unregulated_fixed_port_id: Do not reject unregulated fixed port identifiers.
        As demanded by the specification, the frontend rejects unregulated fixed port ID by default.
        This is a dangerous feature that must not be used unless you understand the risks.
        Please read https://opencyphal.org/guide.

    :return: A Tuple of lists of :class:`pydsdl.CompositeType`. The first index in the Tuple are the types parsed from
        the ``dsdl_files`` argument. The second index are types that the target ``dsdl_files`` utilizes.
        A note for using these values to describe build dependencies: each :class:`pydsdl.CompositeType` has two
        fields that provide links back to the filesystem where the dsdl files were located when parsing the type;
        ``source_file_path`` and ``source_file_path_to_root``.

    :raises: :class:`pydsdl.FrontendError`, :class:`MemoryError`, :class:`SystemError`,
        :class:`OSError` if directories do not exist or inaccessible,
        :class:`ValueError`/:class:`TypeError` if the arguments are invalid.
    """
    # Normalize paths and remove duplicates. Resolve symlinks to avoid ambiguities.
    normal_root_namespace_directories_or_names = normalize_paths_argument_to_list(root_namespace_directories_or_names)
    target_dsdl_definitions = _construct_dsdl_definitions_from_files(
        normalize_paths_argument_to_list(dsdl_files),
        normal_root_namespace_directories_or_names,
    )
    if len(target_dsdl_definitions) == 0:
        _logger.info("No DSDL files found in the specified directories")
        return ([], [])

    if _logger.isEnabledFor(logging.DEBUG):  # pragma: no cover
        _logger.debug("Target DSDL definitions are listed below:")

        for x in target_dsdl_definitions:
            _logger.debug(_LOG_LIST_ITEM_PREFIX + str(x.file_path))

    root_namespaces = {f.root_namespace_path.resolve() for f in target_dsdl_definitions}
    lookup_directories_path_list = _construct_lookup_directories_path_list(
        itertools.chain(root_namespaces, filter(lambda x: x.exists(), normal_root_namespace_directories_or_names)),
        normalize_paths_argument_to_list(lookup_directories),
        True,
    )

    definitions = _complete_read_function(
        target_dsdl_definitions, lookup_directories_path_list, print_output_handler, allow_unregulated_fixed_port_id
    )

    return (definitions.direct, definitions.transitive)


# +--[INTERNAL API::PUBLIC API HELPERS]-------------------------------------------------------------------------------+
# These are functions called by the public API before the actual processing begins.

DSDL_FILE_SUFFIX = ".dsdl"
DSDL_FILE_GLOB = f"*{DSDL_FILE_SUFFIX}"
DSDL_FILE_SUFFIX_LEGACY = ".uavcan"
DSDL_FILE_GLOB_LEGACY = f"*{DSDL_FILE_SUFFIX_LEGACY}"
_LOG_LIST_ITEM_PREFIX = " " * 4


def _complete_read_function(
    target_dsdl_definitions: SortedFileList[ReadableDSDLFile],
    lookup_directories_path_list: list[Path],
    print_output_handler: PrintOutputHandler | None,
    allow_unregulated_fixed_port_id: bool,
) -> DSDLDefinitions:

    lookup_dsdl_definitions = _construct_dsdl_definitions_from_namespaces(lookup_directories_path_list)

    _logger.debug("Lookup DSDL definitions are listed below:")
    for x in lookup_dsdl_definitions:
        _logger.debug(_LOG_LIST_ITEM_PREFIX + str(x))

    _logger.info(
        "Reading %d definitions from the root namespace %s, "
        "with %d lookup definitions located in root namespaces: %s",
        len(target_dsdl_definitions),
        list(set(map(lambda t: t.root_namespace, target_dsdl_definitions)))[0],
        len(lookup_dsdl_definitions),
        ", ".join(set(sorted(map(lambda t: t.root_namespace, lookup_dsdl_definitions)))),
    )

    # This is the biggie. All the rest of the wrangling is just to get to this point. This will take the
    # most time and memory.
    definitions = read_definitions(
        target_dsdl_definitions, lookup_dsdl_definitions, print_output_handler, allow_unregulated_fixed_port_id
    )

    # Note that we check for collisions in the read namespace only.
    # We intentionally ignore (do not check for) possible collisions in the lookup directories,
    # because that would exceed the expected scope of responsibility of the frontend, and the lookup
    # directories may contain issues and mistakes that are outside of the control of the user (e.g.,
    # they could be managed by a third party) -- the user shouldn't be affected by mistakes committed
    # by the third party.
    _ensure_no_fixed_port_id_collisions(definitions.direct)
    _ensure_minor_version_compatibility(definitions.transitive + definitions.direct)

    return definitions


def _construct_lookup_directories_path_list(
    root_namespace_directories: Iterable[Path],
    lookup_directories_path_list: list[Path],
    allow_root_namespace_name_collision: bool,
) -> list[Path]:
    """
    Intermediate transformation and validation of inputs into a list of lookup directories as paths.

    :param root_namespace_directories: The path of the root namespace directory that will be read.
        For example, ``dsdl/uavcan`` to read the ``uavcan`` namespace.

    :param lookup_directories_path_list: List of other namespace directories containing data type definitions that are
        referred to from the target root namespace. For example, if you are reading a vendor-specific namespace,
        the list of lookup directories should always include a path to the standard root namespace ``uavcan``,
        otherwise the types defined in the vendor-specific namespace won't be able to use data types from the
        standard namespace.

    :param allow_root_namespace_name_collision: Allow using the source root namespace name in the look up dirs or
             the same root namespace name multiple times in the lookup dirs. This will enable defining a namespace
             partially and let other entities define new messages or new sub-namespaces in the same root namespace.

    :return: A list of lookup directories as paths.

    :raises: :class:`pydsdl.FrontendError`, :class:`MemoryError`, :class:`SystemError`,
        :class:`OSError` if directories do not exist or inaccessible,
        :class:`ValueError`/:class:`TypeError` if the arguments are invalid.
    """
    # Add the own root namespace to the set of lookup directories, sort lexicographically, remove duplicates.
    # We'd like this to be an iterable list of strings but we handle the common practice of passing in a single path.

    # Normalize paths and remove duplicates. Resolve symlinks to avoid ambiguities.
    lookup_directories_path_list.extend(root_namespace_directories)
    lookup_directories_path_list = list(sorted({x.resolve() for x in lookup_directories_path_list}))
    _logger.debug("Lookup directories are listed below:")
    for a in lookup_directories_path_list:
        _logger.debug(_LOG_LIST_ITEM_PREFIX + str(a))

    # Check for common usage errors and warn the user if anything looks suspicious.
    _ensure_no_common_usage_errors(root_namespace_directories, lookup_directories_path_list, _logger.warning)

    # Check the namespaces and ensure that there are no name collisions.
    _ensure_no_namespace_name_collisions_or_nested_root_namespaces(
        lookup_directories_path_list, allow_root_namespace_name_collision
    )

    return lookup_directories_path_list


def _construct_dsdl_definitions_from_files(
    dsdl_files: list[Path],
    valid_roots: list[Path],
) -> SortedFileList[ReadableDSDLFile]:
    """ """
    output = set()  # type:  set[ReadableDSDLFile]
    for fp in dsdl_files:
        if fp.suffix == DSDL_FILE_SUFFIX_LEGACY:
            _logger.warning(
                "File uses deprecated extension %r, please rename to use %r: %s",
                DSDL_FILE_SUFFIX_LEGACY,
                DSDL_FILE_SUFFIX,
                fp,
            )
        output.add(_dsdl_definition.DSDLDefinition.from_first_in(fp, list(valid_roots)))

    return dsdl_file_sort(output)


def _construct_dsdl_definitions_from_namespaces(
    root_namespace_paths: list[Path],
) -> SortedFileList[ReadableDSDLFile]:
    """
    Accepts a directory path, returns a sorted list of abstract DSDL file representations. Those can be read later.
    The definitions are sorted by name lexicographically, then by major version (greatest version first),
    then by minor version (same ordering as the major version).
    """
    source_file_paths: set[tuple[Path, Path]] = set()  # index of all file paths already found
    for root_namespace_path in root_namespace_paths:
        for p in root_namespace_path.rglob(DSDL_FILE_GLOB):
            source_file_paths.add((p, root_namespace_path))
        for p in root_namespace_path.rglob(DSDL_FILE_GLOB_LEGACY):
            source_file_paths.add((p, root_namespace_path))
            _logger.warning(
                "File uses deprecated extension %r, please rename to use %r: %s",
                DSDL_FILE_GLOB_LEGACY,
                DSDL_FILE_GLOB,
                p,
            )

    return dsdl_file_sort([_dsdl_definition.DSDLDefinition(*p) for p in source_file_paths])


def _ensure_no_fixed_port_id_collisions(types: list[_serializable.CompositeType]) -> None:
    for a in types:
        for b in types:
            different_names = a.full_name != b.full_name
            different_major_versions = a.version.major != b.version.major
            # Must be the same kind because port ID sets of subjects and services are orthogonal
            same_kind = isinstance(a, _serializable.ServiceType) == isinstance(b, _serializable.ServiceType)
            # Data types where the major version is zero are allowed to collide
            both_released = (a.version.major > 0) and (b.version.major > 0)

            fpid_must_be_different = same_kind and (different_names or (different_major_versions and both_released))

            if fpid_must_be_different:
                if a.has_fixed_port_id and b.has_fixed_port_id:
                    if a.fixed_port_id == b.fixed_port_id:
                        raise FixedPortIDCollisionError(
                            "The fixed port ID of this definition is also used in %s" % b.source_file_path,
                            path=a.source_file_path,
                        )


def _ensure_minor_version_compatibility(types: list[_serializable.CompositeType]) -> None:
    by_name = collections.defaultdict(list)  # type: DefaultDict[str, list[_serializable.CompositeType]]
    for t in types:
        by_name[t.full_name].append(t)

    for definitions in by_name.values():
        by_major = collections.defaultdict(list)  # type: DefaultDict[int, list[_serializable.CompositeType]]
        for t in definitions:
            by_major[t.version.major].append(t)

        for subject_to_check in by_major.values():
            _logger.debug("Minor version compatibility check amongst: %s", [str(x) for x in subject_to_check])
            for a in subject_to_check:
                for b in subject_to_check:
                    if a is not b:
                        _ensure_minor_version_compatibility_pairwise(a, b)


def _ensure_minor_version_compatibility_pairwise(
    a: _serializable.CompositeType, b: _serializable.CompositeType
) -> None:
    assert a is not b
    assert a.full_name == b.full_name
    assert a.version.major == b.version.major
    assert a.version.minor != b.version.minor  # This is the whole point of this function.

    # Must be of the same kind: both messages or both services
    if isinstance(a, _serializable.ServiceType) != isinstance(b, _serializable.ServiceType):
        raise VersionsOfDifferentKindError(
            "This definition is not of the same kind as %s" % b.source_file_path, path=a.source_file_path
        )

    # Must use either the same RPID, or the older one should not have an RPID
    if a.has_fixed_port_id == b.has_fixed_port_id:
        if a.fixed_port_id != b.fixed_port_id:
            raise MinorVersionFixedPortIDError(
                "Different fixed port ID values under the same version %s" % b.source_file_path, path=a.source_file_path
            )
    else:
        must_have = a if a.version.minor > b.version.minor else b
        if not must_have.has_fixed_port_id:
            raise MinorVersionFixedPortIDError(
                "Fixed port ID cannot be removed under the same major version", path=must_have.source_file_path
            )

    # Extent and sealing equality
    if isinstance(a, _serializable.ServiceType) and isinstance(b, _serializable.ServiceType):
        _ensure_minor_version_compatibility_pairwise(a.request_type, b.request_type)
        _ensure_minor_version_compatibility_pairwise(a.response_type, b.response_type)
    elif a.version.major > 0:  # Types with major=0 are exempt from compatibility requirements.
        if a.extent != b.extent:
            raise ExtentConsistencyError(
                "The extent of %s is %d bits, whereas the extent of %s is %d bits. "
                "The types share the same major version, so their extents should be equal "
                "to avoid wire compatibility issues." % (a, a.extent, b, b.extent),
                path=a.source_file_path,
            )
        a_sealed = not isinstance(a, _serializable.DelimitedType)
        b_sealed = not isinstance(b, _serializable.DelimitedType)
        sealing_name = ["delimited", "sealed"]
        if a_sealed != b_sealed:
            raise SealingConsistencyError(
                "%s is %s, but %s is %s. "
                "Mixing sealed and delimited types under the same major version will cause wire compatibility issues."
                % (
                    a,
                    sealing_name[a_sealed],
                    b,
                    sealing_name[b_sealed],
                ),
                path=a.source_file_path,
            )


def _ensure_no_common_usage_errors(
    root_namespace_directories: Iterable[Path], lookup_directories: Iterable[Path], reporter: Callable[[str], None]
) -> None:
    suspicious_base_names = [
        "public_regulated_data_types",
        "dsdl",
    ]

    def is_valid_name(s: str) -> bool:
        try:
            _serializable.check_name(s)
        except _error.InvalidDefinitionError:
            return False
        else:
            return True

    # resolve() will also normalize the case in case-insensitive filesystems.
    all_paths = {y.resolve() for y in root_namespace_directories} | {x.resolve() for x in lookup_directories}
    for p in all_paths:
        try:
            candidates = [x for x in p.iterdir() if x.is_dir() and is_valid_name(x.name)]
        except OSError:  # pragma: no cover
            candidates = []
        if candidates and p.name in suspicious_base_names:
            report = (
                "Possibly incorrect usage detected: input path %s is likely incorrect because the last path component "
                "should be the root namespace name rather than its parent directory. You probably meant:\n%s"
            ) % (
                p,
                "\n".join(("- %s" % (p / s)) for s in candidates),
            )
            reporter(report)


def _ensure_no_namespace_name_collisions_or_nested_root_namespaces(
    directories: Iterable[Path], allow_name_collisions: bool
) -> None:
    directories = {x.resolve() for x in directories}  # normalize the case in case-insensitive filesystems

    def check_each(path_tuple_with_result: tuple[tuple[Path, Path], list[int]]) -> bool:
        path_tuple = path_tuple_with_result[0]
        if not path_tuple[0].samefile(path_tuple[1]):
            if not allow_name_collisions and path_tuple[0].name.lower() == path_tuple[1].name.lower():
                return True
            try:
                path_tuple[0].relative_to(path_tuple[1])
            except ValueError:
                pass
            else:
                path_tuple_with_result[1][0] = 1
                return True
        return False

    # zip a list[1] of int 0 so we can assign a failure type. 0 is name collision and 1 is nested root namespace
    # further cartesian checks can be added here using this pattern

    # next/filter returns the first failure or None if no failures
    check_result = next(filter(check_each, zip(product(directories, directories), repeat([0]))), None)

    if check_result:
        path_tuple = check_result[0]
        failure_type = check_result[1][0]
        if failure_type == 0:
            raise RootNamespaceNameCollisionError(
                "The following namespaces have the same name: %s" % path_tuple[0], path=path_tuple[1]
            )
        else:
            raise NestedRootNamespaceError(
                "The following namespace is nested inside this one, which is not permitted: %s" % path_tuple[0],
                path=path_tuple[1],
            )


# +--[ UNIT TESTS ]---------------------------------------------------------------------------------------------------+


def _unittest_dsdl_definition_constructor() -> None:
    import tempfile

    from ._dsdl_definition import FileNameFormatError

    with tempfile.TemporaryDirectory() as directory:
        di = Path(directory).resolve()
        root = di / "foo"
        (root / "nested").mkdir(parents=True)

        (root / "123.Qwerty.123.234.dsdl").write_text("# TEST A")
        (root / "nested/2.Asd.21.32.dsdl").write_text("# TEST B")
        (root / "nested/Foo.32.43.dsdl").write_text("# TEST C")

        dsdl_defs = _construct_dsdl_definitions_from_namespaces([root])
        print(dsdl_defs)
        lut = {x.full_name: x for x in dsdl_defs}  # type: dict[str, ReadableDSDLFile]
        assert len(lut) == 3

        assert str(lut["foo.Qwerty"]) == repr(lut["foo.Qwerty"])
        assert (
            str(lut["foo.Qwerty"])
            == "DSDLDefinition(full_name='foo.Qwerty', version=Version(major=123, minor=234), fixed_port_id=123, "
            "file_path=%s)" % lut["foo.Qwerty"].file_path
        )

        assert (
            str(lut["foo.nested.Foo"])
            == "DSDLDefinition(full_name='foo.nested.Foo', version=Version(major=32, minor=43), fixed_port_id=None, "
            "file_path=%s)" % lut["foo.nested.Foo"].file_path
        )

        t = lut["foo.Qwerty"]
        assert t.file_path == root / "123.Qwerty.123.234.dsdl"
        assert t.has_fixed_port_id
        assert t.fixed_port_id == 123
        assert t.text == "# TEST A"
        assert t.version.major == 123
        assert t.version.minor == 234
        assert t.name_components == ["foo", "Qwerty"]
        assert t.short_name == "Qwerty"
        assert t.root_namespace == "foo"
        assert t.full_namespace == "foo"

        t = lut["foo.nested.Asd"]
        assert t.file_path == root / "nested" / "2.Asd.21.32.dsdl"
        assert t.has_fixed_port_id
        assert t.fixed_port_id == 2
        assert t.text == "# TEST B"
        assert t.version.major == 21
        assert t.version.minor == 32
        assert t.name_components == ["foo", "nested", "Asd"]
        assert t.short_name == "Asd"
        assert t.root_namespace == "foo"
        assert t.full_namespace == "foo.nested"

        t = lut["foo.nested.Foo"]
        assert t.file_path == root / "nested" / "Foo.32.43.dsdl"
        assert not t.has_fixed_port_id
        assert t.fixed_port_id is None
        assert t.text == "# TEST C"
        assert t.version.major == 32
        assert t.version.minor == 43
        assert t.name_components == ["foo", "nested", "Foo"]
        assert t.short_name == "Foo"
        assert t.root_namespace == "foo"
        assert t.full_namespace == "foo.nested"

        (root / "nested/Malformed.MAJOR.MINOR.dsdl").touch()
        try:
            _construct_dsdl_definitions_from_namespaces([root])
        except FileNameFormatError as ex:
            print(ex)
            (root / "nested/Malformed.MAJOR.MINOR.dsdl").unlink()
        else:  # pragma: no cover
            assert False

        (root / "nested/NOT_A_NUMBER.Malformed.1.0.dsdl").touch()
        try:
            _construct_dsdl_definitions_from_namespaces([root])
        except FileNameFormatError as ex:
            print(ex)
            (root / "nested/NOT_A_NUMBER.Malformed.1.0.dsdl").unlink()
        else:  # pragma: no cover
            assert False

        (root / "nested/Malformed.dsdl").touch()
        try:
            _construct_dsdl_definitions_from_namespaces([root])
        except FileNameFormatError as ex:
            print(ex)
            (root / "nested/Malformed.dsdl").unlink()
        else:  # pragma: no cover
            assert False

        _construct_dsdl_definitions_from_namespaces([root])  # making sure all errors are cleared

        (root / "nested/super.bad").mkdir()
        (root / "nested/super.bad/Unreachable.1.0.dsdl").touch()
        try:
            _construct_dsdl_definitions_from_namespaces([root])
        except FileNameFormatError as ex:
            print(ex)
        else:  # pragma: no cover
            assert False

        try:
            _construct_dsdl_definitions_from_namespaces([root / "nested/super.bad"])
        except FileNameFormatError as ex:
            print(ex)
        else:  # pragma: no cover
            assert False

        (root / "nested/super.bad/Unreachable.1.0.dsdl").unlink()


def _unittest_dsdl_definition_constructor_legacy() -> None:
    import tempfile

    with tempfile.TemporaryDirectory() as directory:
        di = Path(directory).resolve()
        root = di / "foo"
        root.mkdir()
        (root / "123.Qwerty.123.234.uavcan").write_text("# TEST A")
        dsdl_defs = _construct_dsdl_definitions_from_namespaces([root])
        print(dsdl_defs)
        lut = {x.full_name: x for x in dsdl_defs}  # type: dict[str, ReadableDSDLFile]
        assert len(lut) == 1
        t = lut["foo.Qwerty"]
        assert t.file_path == root / "123.Qwerty.123.234.uavcan"
        assert t.has_fixed_port_id
        assert t.fixed_port_id == 123
        assert t.text == "# TEST A"
        assert t.version.major == 123
        assert t.version.minor == 234
        assert t.name_components == ["foo", "Qwerty"]
        assert t.short_name == "Qwerty"
        assert t.root_namespace == "foo"
        assert t.full_namespace == "foo"


def _unittest_common_usage_errors() -> None:
    import tempfile

    with tempfile.TemporaryDirectory() as directory:
        di = Path(directory)
        root_ns_dir = di / "foo"
        root_ns_dir.mkdir()

        reports = []  # type: list[str]

        _ensure_no_common_usage_errors([root_ns_dir], [], reports.append)
        assert not reports
        _ensure_no_common_usage_errors([root_ns_dir], [di / "baz"], reports.append)
        assert not reports

        dir_dsdl = root_ns_dir / "dsdl"
        dir_dsdl.mkdir()
        _ensure_no_common_usage_errors([dir_dsdl], [di / "baz"], reports.append)
        assert not reports  # Because empty.

        dir_dsdl_vscode = dir_dsdl / ".vscode"
        dir_dsdl_vscode.mkdir()
        _ensure_no_common_usage_errors([dir_dsdl], [di / "baz"], reports.append)
        assert not reports  # Because the name is not valid.

        dir_dsdl_uavcan = dir_dsdl / "uavcan"
        dir_dsdl_uavcan.mkdir()
        _ensure_no_common_usage_errors([dir_dsdl], [di / "baz"], reports.append)
        (rep,) = reports
        reports.clear()
        assert str(dir_dsdl_uavcan.resolve()).lower() in rep.lower()


def _unittest_nested_roots() -> None:
    import tempfile

    from pytest import raises

    with tempfile.TemporaryDirectory() as directory:
        di = Path(directory)
        (di / "a").mkdir()
        (di / "aa").mkdir()
        (di / "a/b").mkdir()
        (di / "a/c").mkdir()
        (di / "aa/b").mkdir()
        _ensure_no_namespace_name_collisions_or_nested_root_namespaces([], True)
        _ensure_no_namespace_name_collisions_or_nested_root_namespaces([di / "a"], True)
        _ensure_no_namespace_name_collisions_or_nested_root_namespaces([di / "a/b", di / "a/c"], True)
        with raises(NestedRootNamespaceError):
            _ensure_no_namespace_name_collisions_or_nested_root_namespaces([di / "a/b", di / "a"], True)
        _ensure_no_namespace_name_collisions_or_nested_root_namespaces([di / "aa/b", di / "a"], True)
        _ensure_no_namespace_name_collisions_or_nested_root_namespaces([di / "a/b", di / "aa"], True)


def _unittest_issue_71() -> None:  # https://github.com/OpenCyphal/pydsdl/issues/71
    import tempfile

    with tempfile.TemporaryDirectory() as directory:
        real = Path(directory, "real", "nested")
        real.mkdir(parents=True)
        link = Path(directory, "link")
        link.symlink_to(real, target_is_directory=True)
        (real / "Msg.0.1.dsdl").write_text("@sealed")
        assert len(read_namespace(real, [real, link])) == 1
        assert len(read_namespace(link, [real, link])) == 1


def _unittest_type_read_files_example(temp_dsdl_factory) -> None:  # type: ignore
    # let's test the comments for the read function
    dsdl_files = [
        Path("workspace/project/types/animals/felines/Tabby.1.0.uavcan"),  # keep .uavcan to cover the warning
        Path("workspace/project/types/animals/canines/Boxer.1.0.dsdl"),
        Path("workspace/project/types/plants/trees/DouglasFir.1.0.dsdl"),
    ]

    dsdl_files_abs = []
    root_namespace_paths = set()
    for dsdl_file in dsdl_files:
        dsdl_files_abs.append(temp_dsdl_factory.new_file(dsdl_file, "@sealed"))
        root_namespace_paths.add(temp_dsdl_factory.base_dir / dsdl_file.parent.parent)
    root_namespace_directories_or_names_simple = ["animals", "plants"]

    direct, transitive = read_files(dsdl_files_abs, root_namespace_directories_or_names_simple)

    assert len(direct) == len(dsdl_files)
    assert len(transitive) == 0

    for direct_type in direct:
        assert direct_type.root_namespace in root_namespace_directories_or_names_simple
        assert direct_type.source_file_path_to_root in root_namespace_paths

    direct, _ = read_files(dsdl_files_abs, root_namespace_paths)

    assert len(direct) == len(dsdl_files)

    for direct_type in direct:
        assert direct_type.root_namespace in root_namespace_directories_or_names_simple
        assert direct_type.source_file_path_to_root in root_namespace_paths


def _unittest_targets_found_in_lookup_namespaces(temp_dsdl_factory) -> None:  # type: ignore

    # call read_files with a list of dsdl files which are also located in the provided lookup namespaces

    plant_1_0 = Path("types/plants/Plant.1.0.dsdl")
    tree_1_0 = Path("types/plants/trees/Tree.1.0.dsdl")
    douglas_fir_1_0 = Path("types/plants/trees/DouglasFir.1.0.dsdl")

    plant_file = temp_dsdl_factory.new_file(plant_1_0, "@sealed\n")
    test_files = [
        temp_dsdl_factory.new_file(tree_1_0, "@sealed\nplants.Plant.1.0 plt\n"),
        temp_dsdl_factory.new_file(douglas_fir_1_0, "@sealed\nplants.trees.Tree.1.0 tree\n"),
    ]
    lookup_dirs = [plant_file.parent]

    direct, transitive = read_files(test_files, lookup_dirs)

    assert len(direct) == len(test_files)
    assert len(transitive) == 1


def _unittest_read_files_empty_args() -> None:
    direct, transitive = read_files([], [])

    assert len(direct) == 0
    assert len(transitive) == 0


def _unittest_ensure_no_namespace_name_collisions_or_nested_root_namespaces() -> None:
    """gratuitous coverage of the collision check where other tests don't cover some edge cases."""
    _ensure_no_namespace_name_collisions_or_nested_root_namespaces([], False)


def _unittest_issue_104(temp_dsdl_factory) -> None:  # type: ignore
    """demonstrate that removing _ensure_no_collisions is okay"""

    from pytest import raises

    thing_1_0 = Path("a/b/thing.1.0.dsdl")
    thing_type_1_0 = Path("a/b/thing/thingtype.1.0.dsdl")

    file_at_root = temp_dsdl_factory.new_file(Path("a/Nothing.1.0.dsdl"), "@sealed\n")
    thing_file = temp_dsdl_factory.new_file(thing_1_0, "@sealed\na.b.thing.thingtype.1.0 thing\n")
    _ = temp_dsdl_factory.new_file(thing_type_1_0, "@sealed\n")

    direct, transitive = read_files(thing_file, file_at_root.parent, file_at_root.parent)

    assert len(direct) == 1
    assert len(transitive) == 1

    thing_1_1 = Path("a/b/thing.1.1.dsdl")

    thing_file2 = temp_dsdl_factory.new_file(thing_1_1, "@sealed\na.b.thing.Thingtype.1.0 thing\n")

    from ._data_type_builder import DataTypeNameCollisionError

    with raises(DataTypeNameCollisionError):
        read_files(thing_file2, file_at_root.parent, file_at_root.parent)


def _unittest_issue_109(temp_dsdl_factory) -> None:  # type: ignore
    """Allow target paths to be relative"""

    # pylint: disable=too-many-locals

    from pytest import raises
    from ._dsdl_definition import PathInferenceError

    odin_1_0 = Path("gods/norse/Odin.1.0.dsdl")
    huginn_1_0 = Path("familiars/norse/Huginn.1.0.dsdl")
    muninn_1_0 = Path("familiars/norse/Muninn.1.0.dsdl")

    odin_file = temp_dsdl_factory.new_file(
        odin_1_0, "@sealed\nfamiliars.norse.Huginn.1.0 huginn\nfamiliars.norse.Muninn.1.0 muninn\n"
    )
    huginn_file = temp_dsdl_factory.new_file(huginn_1_0, "@sealed\n")
    _ = temp_dsdl_factory.new_file(muninn_1_0, "@sealed\n")

    # relative path for target / absolute path for root namespace
    direct1, transitive1 = read_files([odin_1_0], [odin_file.parent.parent, huginn_file.parent.parent])

    assert len(direct1) == 1
    assert len(transitive1) == 2

    assert direct1[0].source_file_path == odin_file

    # here we use an absolute path to the target and provide a root namespace name for it.
    direct2, transitive2 = read_files([odin_file], ["gods"], [huginn_file.parent.parent])

    assert len(direct2) == 1
    assert len(transitive2) == 2

    assert direct2[0].source_file_path == odin_file

    # Once more to show that it works if the name and lookup is in the second argument
    direct3, transitive3 = read_files([odin_file], ["gods", huginn_file.parent.parent])

    assert len(direct3) == 1
    assert len(transitive3) == 2

    assert direct3[0].source_file_path == odin_file

    # Next show that the lookup paths cannot be used to find targets
    with raises(PathInferenceError):
        _ = read_files([odin_1_0], [], [odin_file.parent.parent, huginn_file.parent.parent])

    # Finally, we'll make sure that we don't confuse roots when two targets are rooted in the same place
    # Once more to show that it works if the name and lookup is in the second argument
    direct4, transitive4 = read_files([odin_1_0], [huginn_file.parent.parent, odin_file.parent.parent])

    assert len(direct4) == 1
    assert len(transitive4) == 2

    assert direct4[0].source_file_path == odin_file
    assert direct4[0].root_namespace == "gods"



# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT

from __future__ import annotations
import functools
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import cast

from ._dsdl import DefinitionVisitor, DSDLFile, ReadableDSDLFile, PrintOutputHandler, SortedFileList
from ._dsdl import file_sort as dsdl_file_sort
from ._error import FrontendError, InternalError
from ._serializable._composite import CompositeType


# pylint: disable=too-many-arguments
def _read_definitions(
    target_definitions: SortedFileList[ReadableDSDLFile],
    lookup_definitions: SortedFileList[ReadableDSDLFile],
    print_output_handler: PrintOutputHandler | None,
    allow_unregulated_fixed_port_id: bool,
    direct: set[CompositeType],
    transitive: set[CompositeType],
    file_pool: dict[Path, ReadableDSDLFile],
    level: int,
) -> None:
    """
    Don't look at me! I'm hideous!
    (recursive method with a lot of arguments. See read_definitions for documentation)
    """

    _pending_definitions: set[ReadableDSDLFile] = set()

    class _Callback(DefinitionVisitor):
        def on_definition(self, _: DSDLFile, dependency_dsdl_file: ReadableDSDLFile) -> None:
            if dependency_dsdl_file.file_path not in file_pool:
                _pending_definitions.add(dependency_dsdl_file)

    def print_handler(file: Path, line: int, message: str) -> None:
        if print_output_handler is not None:
            print_output_handler(file, line, message)

    for target_definition in target_definitions:

        if not isinstance(target_definition, ReadableDSDLFile):
            raise TypeError("Expected ReadableDSDLFile, got: " + type(target_definition).__name__)

        target_definition = file_pool.setdefault(target_definition.file_path, target_definition)
        # make sure we are working with the same object for a given file path

        if target_definition.composite_type is not None and (
            target_definition.composite_type in direct or target_definition.composite_type in transitive
        ):
            logging.debug("Skipping target file %s because it has already been processed", target_definition.file_path)
            if level == 0 and target_definition.composite_type in transitive:
                # promote to direct
                transitive.remove(target_definition.composite_type)
                direct.add(target_definition.composite_type)
            continue

        try:
            new_composite_type = target_definition.read(
                lookup_definitions,
                [_Callback()],
                functools.partial(print_handler, target_definition.file_path),
                allow_unregulated_fixed_port_id,
            )
        except FrontendError as ex:  # pragma: no cover
            ex.set_error_location_if_unknown(path=target_definition.file_path)
            raise ex
        except Exception as ex:  # pragma: no cover
            raise InternalError(culprit=ex, path=target_definition.file_path) from ex

        if level == 0:

            direct.add(new_composite_type)
            try:
                transitive.remove(new_composite_type)
            except KeyError:
                pass
        else:
            transitive.add(new_composite_type)

        if len(_pending_definitions) > 0:
            _read_definitions(
                dsdl_file_sort(_pending_definitions),
                lookup_definitions,
                print_output_handler,
                allow_unregulated_fixed_port_id,
                direct,
                transitive,
                file_pool,
                level + 1,
            )
            _pending_definitions.clear()


# +---[FILE: PUBLIC]--------------------------------------------------------------------------------------------------+


@dataclass(frozen=True)
class DSDLDefinitions:
    """
    Common DSDL definition set including the direct dependencies requested and the transitive dependencies found.
    The former and latter sets will be disjoint.
    """

    direct: SortedFileList[CompositeType]
    transitive: SortedFileList[CompositeType]


def read_definitions(
    target_definitions: SortedFileList[ReadableDSDLFile],
    lookup_definitions: SortedFileList[ReadableDSDLFile],
    print_output_handler: PrintOutputHandler | None,
    allow_unregulated_fixed_port_id: bool,
) -> DSDLDefinitions:
    """
    Given a set of DSDL files, this method reads the text and invokes the parser for each and for any files found in the
    lookup set where these are used by the target set.

    :param target_definitions:              List of definitions to read.
    :param lookup_definitions:              List of definitions available for referring to.
    :param print_output_handler:            Used for @print and for diagnostics: (line_number, text) -> None.
    :param allow_unregulated_fixed_port_id: Do not complain about fixed unregulated port IDs.
    :return: The data type representation.
    :raises InvalidDefinitionError: If a dependency is missing.
    :raises InternalError: If an unexpected error occurs.
    """
    _direct: set[CompositeType] = set()
    _transitive: set[CompositeType] = set()
    _file_pool: dict[Path, ReadableDSDLFile] = {}
    _read_definitions(
        target_definitions,
        lookup_definitions,
        print_output_handler,
        allow_unregulated_fixed_port_id,
        _direct,
        _transitive,
        _file_pool,
        0,
    )
    return DSDLDefinitions(
        dsdl_file_sort(_direct),
        dsdl_file_sort(_transitive),
    )


# +-[UNIT TESTS]------------------------------------------------------------------------------------------------------+


def _unittest_namespace_reader_read_definitions(temp_dsdl_factory) -> None:  # type: ignore
    from . import _dsdl_definition

    target = temp_dsdl_factory.new_file(Path("root", "ns", "Target.1.1.dsdl"), "@sealed")
    target_definitions = [cast(ReadableDSDLFile, _dsdl_definition.DSDLDefinition(target, target.parent))]
    lookup_definitions: list[ReadableDSDLFile] = []

    read_definitions(target_definitions, lookup_definitions, None, True)


def _unittest_namespace_reader_read_definitions_multiple(temp_dsdl_factory) -> None:  # type: ignore
    from . import _dsdl_definition

    targets = [
        temp_dsdl_factory.new_file(Path("root", "ns", "Target.1.1.dsdl"), "@sealed\nns.Aisle.1.0 paper_goods\n"),
        temp_dsdl_factory.new_file(Path("root", "ns", "Target.2.0.dsdl"), "@sealed\nns.Aisle.2.0 paper_goods\n"),
        temp_dsdl_factory.new_file(Path("root", "ns", "Walmart.2.4.dsdl"), "@sealed\nns.Aisle.1.0 paper_goods\n"),
    ]
    aisles = [
        temp_dsdl_factory.new_file(Path("root", "ns", "Aisle.1.0.dsdl"), "@sealed"),
        temp_dsdl_factory.new_file(Path("root", "ns", "Aisle.2.0.dsdl"), "@sealed"),
        temp_dsdl_factory.new_file(Path("root", "ns", "Aisle.3.0.dsdl"), "@sealed"),
    ]

    definitions = read_definitions(
        [_dsdl_definition.DSDLDefinition(t, t.parent) for t in targets],
        [_dsdl_definition.DSDLDefinition(a, a.parent) for a in aisles],
        None,
        True,
    )

    assert len(definitions.direct) == 3
    assert len(definitions.transitive) == 2


def _unittest_namespace_reader_read_definitions_multiple_no_load(temp_dsdl_factory) -> None:  # type: ignore
    """
    Ensure that the loader does not load files that are not in the transitive closure of the target files.
    """
    from . import _dsdl_definition

    targets = [
        temp_dsdl_factory.new_file(Path("root", "ns", "Adams.1.0.dsdl"), "@sealed\nns.Tacoma.1.0 volcano\n"),
        temp_dsdl_factory.new_file(Path("root", "ns", "Hood.1.0.dsdl"), "@sealed\nns.Rainer.1.0 volcano\n"),
        temp_dsdl_factory.new_file(Path("root", "ns", "StHelens.2.1.dsdl"), "@sealed\nns.Baker.1.0 volcano\n"),
    ]
    dependencies = [
        temp_dsdl_factory.new_file(Path("root", "ns", "Tacoma.1.0.dsdl"), "@sealed"),
        temp_dsdl_factory.new_file(Path("root", "ns", "Rainer.1.0.dsdl"), "@sealed"),
        temp_dsdl_factory.new_file(Path("root", "ns", "Baker.1.0.dsdl"), "@sealed"),
        temp_dsdl_factory.new_file(Path("root", "ns", "Shasta.1.0.dsdl"), "@sealed"),
        # never be read thus it will not be an error that it does not exist.
    ]

    target_definitions = [cast(ReadableDSDLFile, _dsdl_definition.DSDLDefinition(t, t.parent)) for t in targets]
    lookup_definitions = [cast(ReadableDSDLFile, _dsdl_definition.DSDLDefinition(a, a.parent)) for a in dependencies]
    _ = read_definitions(
        target_definitions,
        lookup_definitions,
        None,
        True,
    )

    # make sure Shasta.1.0 was never accessed but Tacoma 1.0 was
    last_item = lookup_definitions[-1]
    assert isinstance(last_item, _dsdl_definition.DSDLDefinition)
    assert last_item._text is None  # pylint: disable=protected-access
    assert lookup_definitions[0].composite_type is not None

    # Make sure text is cached.
    assert lookup_definitions[0].text == lookup_definitions[0].text


def _unittest_namespace_reader_read_definitions_promotion(temp_dsdl_factory) -> None:  # type: ignore
    from . import _dsdl_definition

    user_1_0 = temp_dsdl_factory.new_file(Path("root", "ns", "User.1.0.dsdl"), "@sealed\n")
    targets = [
        temp_dsdl_factory.new_file(Path("root", "ns", "User.2.0.dsdl"), "@sealed\nns.User.1.0 old_guy\n"),
        user_1_0,
    ]
    lookups = [user_1_0]

    definitions = read_definitions(
        [_dsdl_definition.DSDLDefinition(t, t.parent) for t in targets],
        [_dsdl_definition.DSDLDefinition(l, l.parent) for l in lookups],
        None,
        True,
    )

    assert len(definitions.direct) == 2
    assert len(definitions.transitive) == 0


def _unittest_namespace_reader_read_definitions_no_demote(temp_dsdl_factory) -> None:  # type: ignore
    from . import _dsdl_definition

    user_1_0 = temp_dsdl_factory.new_file(Path("root", "ns", "User.1.0.dsdl"), "@sealed\n")
    targets = [
        user_1_0,
        temp_dsdl_factory.new_file(Path("root", "ns", "User.2.0.dsdl"), "@sealed\nns.User.1.0 old_guy\n"),
    ]
    lookups = [user_1_0]

    definitions = read_definitions(
        [_dsdl_definition.DSDLDefinition(t, t.parent) for t in targets],
        [_dsdl_definition.DSDLDefinition(l, l.parent) for l in lookups],
        None,
        True,
    )

    assert len(definitions.direct) == 2
    assert len(definitions.transitive) == 0


def _unittest_namespace_reader_read_definitions_no_promote(temp_dsdl_factory) -> None:  # type: ignore
    from . import _dsdl_definition

    targets = [
        temp_dsdl_factory.new_file(Path("root", "ns", "User.2.0.dsdl"), "@sealed\nns.User.1.0 old_guy\n"),
        temp_dsdl_factory.new_file(Path("root", "ns", "User.3.0.dsdl"), "@sealed\nns.User.1.0 old_guy\n"),
    ]
    lookups = [temp_dsdl_factory.new_file(Path("root", "ns", "User.1.0.dsdl"), "@sealed\n")]

    definitions = read_definitions(
        [_dsdl_definition.DSDLDefinition(t, t.parent) for t in targets],
        [_dsdl_definition.DSDLDefinition(l, l.parent) for l in lookups],
        None,
        True,
    )

    assert len(definitions.direct) == 2
    assert len(definitions.transitive) == 1


def _unittest_namespace_reader_read_definitions_twice(temp_dsdl_factory) -> None:  # type: ignore
    from . import _dsdl_definition

    targets = [
        temp_dsdl_factory.new_file(Path("root", "ns", "User.2.0.dsdl"), "@sealed\nns.User.1.0 old_guy\n"),
        temp_dsdl_factory.new_file(Path("root", "ns", "User.2.0.dsdl"), "@sealed\nns.User.1.0 old_guy\n"),
    ]
    lookups = [temp_dsdl_factory.new_file(Path("root", "ns", "User.1.0.dsdl"), "@sealed\n")]

    definitions = read_definitions(
        [_dsdl_definition.DSDLDefinition(t, t.parent) for t in targets],
        [_dsdl_definition.DSDLDefinition(l, l.parent) for l in lookups],
        None,
        True,
    )

    assert len(definitions.direct) == 1
    assert len(definitions.transitive) == 1


def _unittest_namespace_reader_read_definitions_missing_dependency(temp_dsdl_factory) -> None:  # type: ignore
    """
    Verify that an error is raised when a dependency is missing.
    """
    from pytest import raises as assert_raises

    from . import _dsdl_definition
    from ._data_type_builder import UndefinedDataTypeError

    with assert_raises(UndefinedDataTypeError):
        read_definitions(
            [
                _dsdl_definition.DSDLDefinition(
                    f := temp_dsdl_factory.new_file(
                        Path("root", "ns", "Cat.1.0.dsdl"), "@sealed\nns.Birman.1.0 fluffy\n"
                    ),
                    f.parent,
                )
            ],
            [],
            None,
            True,
        )


def _unittest_namespace_reader_read_definitions_target_in_lookup(temp_dsdl_factory) -> None:  # type: ignore
    """
    Ensure the direct and transitive sets are disjoint.
    """
    from . import _dsdl_definition

    targets = [
        temp_dsdl_factory.new_file(Path("root", "ns", "Ontario.1.0.dsdl"), "@sealed\nns.NewBrunswick.1.0 place\n"),
        temp_dsdl_factory.new_file(Path("root", "ns", "NewBrunswick.1.0.dsdl"), "@sealed"),
    ]
    lookup = [
        temp_dsdl_factory.new_file(Path("root", "ns", "NewBrunswick.1.0.dsdl"), "@sealed"),
    ]

    definitions = read_definitions(
        [_dsdl_definition.DSDLDefinition(t, t.parent) for t in targets],
        [_dsdl_definition.DSDLDefinition(l, l.parent) for l in lookup],
        None,
        True,
    )

    assert len(definitions.direct) == 2
    assert len(definitions.transitive) == 0


def _unittest_namespace_reader_read_defs_target_dont_allow_unregulated(temp_dsdl_factory) -> None:  # type: ignore
    """
    Ensure that an error is raised when an invalid, fixed port ID is used without an override.
    """
    from pytest import raises as assert_raises

    from . import _dsdl_definition
    from ._data_type_builder import UnregulatedFixedPortIDError

    targets = [
        temp_dsdl_factory.new_file(Path("root", "ns", "845.Lice.1.0.dsdl"), "@sealed\n"),
    ]

    with assert_raises(UnregulatedFixedPortIDError):
        read_definitions(
            [_dsdl_definition.DSDLDefinition(t, t.parent) for t in targets],
            [],
            None,
            False,
        )


def _unittest_namespace_reader_type_error() -> None:
    from pytest import raises as assert_raises

    with assert_raises(TypeError):
        read_definitions(
            [""],  # type: ignore
            [],
            None,
            True,
        )
